import warnings
import numpy as np
import pandas as pd
import nptyping as npt
import xarray as xr
import dask.array as da
from nptyping import NDArray, Int8, Int64, Float64, Bool, Shape, Float, Int
from typing import Any, Hashable, List, Iterable, Callable, Union, Dict, Tuple
from functools import cached_property
from scipy import interpolate
import xftsim as xft


class GeneticMap:
    """
    Map between physical and genetic distances.

    Parameters
    ----------
    chrom : Iterable
        Chromsomes variants are located on
    pos_bp : Iterable
        Physical positions of variants
    pos_cM : Iterable
        Map distances in cM

    Attributes
    __________
    frame : pd.DataFrame
        Pandas DataFrame with the above columns
    chroms : np.ndarray
        Unique chromosomes present in map
    """
    def __init__(self,
                 chrom: Iterable,
                 pos_bp: Iterable,
                 pos_cM: Iterable):
        self.frame = pd.DataFrame.from_dict(dict(chrom = chrom,
                                                 pos_bp = pos_bp,
                                                 pos_cM = pos_cM))
        self.chroms = np.unique(self.frame.chrom.values.astype(int)).astype(str)

    @classmethod
    def from_pyrho_maps(cls, paths: Iterable, sep='\t', **kwargs) -> "GeneticMap":
        """Construct genetic map objects from maps provided at https://github.com/popgenmethods/pyrho
        Please cite their work if you use their maps.
        
        Parameters
        ----------
        paths : Iterable
            Paths for each chromosome
        sep : str, optional
            Passed to pd.read_csv()
        **kwargs
            Additional arguments to pd.read_csv()
        
        Returns
        -------
        GeneticMap
        """
        gmap = pd.concat([pd.read_csv(path, sep='\t', **kwargs) for path in paths])
        chrom = np.char.lstrip(gmap.Chromosome.values.astype(str),'chr')
        pos_bp = gmap['Position(bp)'].values
        pos_cM = gmap['Map(cM)'].values
        return cls(chrom, pos_bp, pos_cM)


    def interpolate_cM_chrom(self, pos_bp: Iterable, chrom: str, **kwargs):
        """
        Interpolate cM values in a specified chromosome based on genetic map information.

        Parameters
        ----------
        pos_bp : Iterable
            Physical positions for which to interpolate cM values
        chrom : str
            Chromosome on which to interpolate
        **kwargs
            Additional keyword arguments to be passed to scipy.interpolate.interp1d.
        """ 
        subset = self.frame[self.frame.chrom==chrom]
        interpolator = interpolate.interp1d(x = subset.pos_bp.values,
                                            y = subset.pos_cM.values,
                                            **kwargs)
        return interpolator(pos_bp)


        if self._col_dim != 'variant':
            raise TypeError
        chroms = np.unique(self._obj.chrom.values.astype(int)).astype(str)
        for chrom in chroms:  
            rmap_chrom = rmap_df[rmap_df['Chromosome']=='chr'+chrom]
            interpolator = interpolate.interp1d(x = rmap_chrom['Position(bp)'].values,
                                                y = rmap_chrom['Map(cM)'].values,
                                                **kwargs)
            self._obj.pos_cM[self._obj.chrom==chrom] = interpolator(self._obj.pos_bp[self._obj.chrom==chrom])

 
                 

@xr.register_dataarray_accessor("xft")
class XftAccessor:
    """
    Accessor for Xarray DataArrays with specialized functionality for HaplotypeArray
    and PhenotypeArray objects.
    
    Parameters
    ----------
    xarray_obj : xarray.DataArray
        The DataArray to be accessed.
    
    Attributes
    ----------
    _obj : xarray.DataArray
        The DataArray to be accessed.
    _array_type : str
        The type of the DataArray, either 'HaplotypeArray' or 'componentArray'.
    _non_annotation_vars : list of str
        The non-annotation variables in the DataArray.
    _variant_vars : list of str
        The variant annotation variables in the DataArray.
    _sample_vars : list of str
        The sample annotation variables in the DataArray.
    _component_vars : list of str
        The component annotation variables in the DataArray.
    _row_dim : str
        The label of the row dimension.
    _col_dim : str
        The label of the column dimension.
    shape : tuple
        The shape of the DataArray.
    n : int
        The number of rows in the DataArray.
    data : numpy.ndarray
        The data in the DataArray.
    row_vars: list
        List of coordinate variable names for the row dimension.
    column_vars: list
        List of coordinate variable names for the column dimension.
    sample_mindex: pd.MultiIndex
        MultiIndex object for the 'sample' dimension, containing iid, fid, and sex columns.
    component_mindex: pd.MultiIndex
        MultiIndex object for the 'component' dimension, containing phenotype_name, component_name, and vorigin_relative columns.


    Raises
    ------
    NotImplementedError
        If the DataArray dimensions are not ('sample', 'variant') or ('sample', 'component').
    """
    
    def __init__(self, xarray_obj):
        self._obj = xarray_obj
        if self._obj.dims == ('sample', 'variant'):
            self._array_type = 'HaplotypeArray'
            self._non_annotation_vars = [
                'variant', 'vid', 'chrom', 'zero_allele', 'one_allele', 'af', 'hcopy', 'pos_bp', 'pos_cM']
            self._variant_vars = ['vid', 'chrom', 'zero_allele', 'one_allele', 'af']
            self._sample_vars = ['iid', 'fid', 'sex']
        elif self._obj.dims == ('sample', 'component'):
            self._array_type = 'componentArray'
            self._component_vars = ['phenotype_name', 'component_name', 'vorigin_relative']
            self._sample_vars = ['iid', 'fid', 'sex']
        else:
            raise NotImplementedError("Unsupported dimensions for DataArray.")
    
    @property
    def _row_dim(self):
        """
        The label of the row dimension.
        
        Returns
        -------
        str
            The label of the row dimension.
        """
        return self._obj.dims[0]

    @property
    def _col_dim(self):
        """
        The label of the column dimension.
        
        Returns
        -------
        str
            The label of the column dimension.
        """
        return self._obj.dims[1]

    @property
    def shape(self):
        """
        The shape of the DataArray.
        
        Returns
        -------
        tuple
            The shape of the DataArray.
        """
        return self._obj.shape

    @property
    def n(self):
        """
        The number of rows in the DataArray.
        
        Returns
        -------
        int
            The number of rows in the DataArray.
        """
        return self._obj.shape[0]

    @property
    def data(self):
        """
        The data in the DataArray.
        
        Returns
        -------
        numpy.ndarray
            The data in the DataArray.
        """
        return self._obj.data

    #### functions for constructing XftIndex objects ####
    def get_sample_indexer(self):
        """
        Returns an instance of `xft.index.SampleIndex` representing the sample indexer
        constructed from the input data.

        Raises
        ------
        NotImplementedError
            If `_row_dim` is not `'sample'`.

        Returns
        -------
        SampleIndex
            An instance of `xft.index.SampleIndex` constructed from the sample data in the
            input object.
        """
        if self._row_dim != 'sample':
            raise NotImplementedError
        return xft.index.SampleIndex(
            iid=self._obj.coords['sample'].iid.data,
            fid=self._obj.coords['sample'].fid.data,
            sex=self._obj.coords['sample'].sex.data,
        )

    def set_sample_indexer(self, value):
        raise NotImplementedError

    def get_variant_indexer(self):
        """
        Get the variant indexer of a HaplotypeArray.

        Returns
        -------
        xft.index.HaploidVariantIndex
            A HaploidVariantIndex object.
        """
        if self._col_dim != 'variant':
            raise NotImplementedError
        annotations = self.get_annotation_dict()
        if len(annotations) == 0:
            annotation_array = None
            annotation_names = None
        else:
            annotation_array, annotation_names = annotations.items()
        return xft.index.HaploidVariantIndex(
            vid=self._obj.coords['variant'].vid.data,
            chrom=self._obj.coords['variant'].chrom.data,
            h_copy=self._obj.coords['variant'].hcopy.data,
            zero_allele=self._obj.coords['variant'].zero_allele.data,
            one_allele=self._obj.coords['variant'].one_allele.data,
            af=self._obj.coords['variant'].af.data,
            pos_bp=self._obj.coords['variant'].pos_bp.data,
            pos_cM=self._obj.coords['variant'].pos_cM.data,
            annotation_array=annotation_array,
            annotation_names=annotation_names)

    def set_variant_indexer(self, value):
        raise NotImplementedError

    def get_component_indexer(self):
        """
        Get the component indexer of a PhenotypeArray.

        Returns
        -------
        xft.index.ComponentIndex
            A ComponentIndex object.
        """
        if self._col_dim != 'component':
            raise NotImplementedError
        return xft.index.ComponentIndex(
            phenotype_name=self._obj.coords['component'].phenotype_name.data,
            component_name=self._obj.coords['component'].component_name.data,
            vorigin_relative=self._obj.coords['component'].vorigin_relative.data,
            comp_type=self._obj.coords['component'].comp_type.data,
        )

    def reindex_components(self, value):
        """
        Reindex the components.

        Parameters
        ----------
        value : xft.index.ComponentIndex
            A ComponentIndex object.

        Returns
        -------
        PhenotypeArray
            A new PhenotypeArray object.

        """
        # ugly as hell, works for now
        return PhenotypeArray(self._obj.data,
                              component_indexer=value,
                              sample_indexer=self.get_sample_indexer(),
                              )
        # self._obj['phenotype_name'] = value.phenotype_name
        # self._obj['component_name'] = value.component_name
        # self._obj['vorigin_relative'] = value.vorigin_relative

    def get_row_indexer(self):
        """
        Get the row indexer.

        Returns
        -------
        xft.index.SampleIndex
            A SampleIndex object.

        Raises
        ------
        TypeError
            If the row dimension is not 'sample'.

        """
        if self._row_dim == 'sample':
            return self.get_sample_indexer()
        else:
            raise TypeError

    def set_row_indexer(self):
        raise NotImplementedError

    def get_column_indexer(self):
        """
        Get the column indexer object for the PhenotypeArray object.

        Returns
        -------
        xft.index.Indexer
            The indexer object based on the current column dimension.

        Raises
        ------
        TypeError
            If the current column dimension is not recognized.
        """
        if self._col_dim == 'variant':
            return self.get_variant_indexer()
        elif self._col_dim == 'component':
            return self.get_component_indexer()
        else:
            raise TypeError

    def set_column_indexer(self, value):
        """
        Set the column indexer object for the PhenotypeArray object.

        Parameters
        ----------
        value : xft.index.Indexer
            The new indexer object for the PhenotypeArray object.

        Returns
        -------
        None

        Raises
        ------
        TypeError
            If the current column dimension is not recognized.
        """
        if self._col_dim == 'variant':
            return self.set_variant_indexer(value)
        elif self._col_dim == 'component':
            return self.set_component_indexer(value)
        else:
            raise TypeError

    @property
    def row_vars(self):
        """
        Get the row coordinate variables for the PhenotypeArray object.

        Returns
        -------
        XftIndex
            The row coordinate variables of the row dimension.
        """
        return self.get_row_indexer()._coord_variables

    @property
    def column_vars(self):
        """
        Get the column coordinate variables for the DataArray object.

        Returns
        -------
        XftIndex
            The column coordinate variables of the current column dimension.
        """
        return self.get_column_indexer()._coord_variables

    # accessors for pd.MultiIndex objects
    @property
    def sample_mindex(self):
        """
        Get the sample multi-index for the PhenotypeArray object.

        Returns
        -------
        pd.MultiIndex
            A multi-index object containing sample IDs, family IDs, and sex information.
        
        Raises
        ------
        NotImplementedError
            If the current row dimension is not 'sample'.
        """
        if self._row_dim != 'sample':
            raise NotImplementedError
        df = pd.DataFrame.from_dict(dict(
            iid=self._obj.coords['sample'].iid.data,
            fid=self._obj.coords['sample'].fid.data,
            sex=self._obj.coords['sample'].sex.data,
        ))
        return pd.MultiIndex.from_frame(df)

    @property
    def component_mindex(self):
        """
        Get a Pandas MultiIndex object for the component dimension.

        Returns
        -------
        pandas.MultiIndex
            MultiIndex object with phenotype_name, component_name, and vorigin_relative
            as index levels.

        Raises
        ------
        NotImplementedError
            If the column dimension is not 'component'.
        """ 
        if self._col_dim != 'component':
            raise NotImplementedError
        df = pd.DataFrame.from_dict(dict(
            phenotype_name=self._obj.coords['component'].phenotype_name.data,
            component_name=self._obj.coords['component'].component_name.data,
            vorigin_relative=self._obj.coords['component'].vorigin_relative.data
        ))
        return pd.MultiIndex.from_frame(df)

    def standardize(self):
        out = self._obj.copy()
        out.data = xft.utils.standardize_array(self._obj.data)
        return out

    ############ HaplotypeArray properties ############

    def interpolate_cM(self,
                       gmap: GeneticMap,
                       # rmap_df: pd.DataFrame = xft.data.get_ceu_map(),
                       **kwargs):
        """
        Interpolate cM values based on genetic map information.
        Specific to HaplotypeArray objects.

        Parameters
        ----------
        gmap : GeneticMap
            Genetic map data
        **kwargs
            Additional keyword arguments to be passed to scipy.interpolate.interp1d.

        Raises
        ------
        TypeError
            If the column dimension is not 'variant'.
        ValueError
            If not all chromosomes required are present in the genetic map
        """
        if self._col_dim != 'variant':
            raise TypeError
        chroms = np.unique(self._obj.chrom.values.astype(int)).astype(str)
        if not (set(chroms) <= set(gmap.chroms)):
            raise ValueError('Not all chromosomes are present on specified genetic Map')
        for chrom in chroms:  
            self._obj.pos_cM[self._obj.chrom==chrom] = gmap.interpolate_cM_chrom(self._obj.pos_bp[self._obj.chrom==chrom], 
                                      chrom=chrom,
                                      **kwargs)
            # self._obj.pos_cM[self._obj.chrom==chrom] = interpolator(self._obj.pos_bp[self._obj.chrom==chrom])

    def use_empirical_afs(self):
        """
        Sets allele frequencies to the empirical frequencies.
        Specific to HaplotypeArray objects.
        
        Raises
        ------
        TypeError
            If `_col_dim` is not 'variant'.
        """
        if self._col_dim != 'variant':
            raise TypeError
        self._obj.af.values = np.repeat(self.af_empirical,2)

    @property
    def diploid_vid(self):
        """
        Diploid variant ID.
        Specific to HaplotypeArray objects.
        
        Returns
        -------
        numpy.ndarray
            Diploid variant IDs.
        
        Raises
        ------
        TypeError
            If `_col_dim` is not 'variant'.
        """
        if self._col_dim != 'variant':
            raise TypeError
        return self._obj.vid[::2]

    @property
    def diploid_chrom(self):
        """
        Diploid chromosome numbers.
        Specific to HaplotypeArray objects.
        
        Returns
        -------
        numpy.ndarray
            Diploid chromosome numbers.
        
        Raises
        ------
        TypeError
            If `_col_dim` is not 'variant'.
        """
        if self._col_dim != 'variant':
            raise TypeError
        return self._obj.chrom[::2]

    @property
    def generation(self):
        """
        Generation of the data.
        Specific to HaplotypeArray objects.
        
        Returns
        -------
        int
            Generation attribute.
        
        Raises
        ------
        TypeError
            If `_col_dim` is not 'variant'.
        """
        if self._col_dim != 'variant':
            raise TypeError
        return self.attrs['generation']

    @property
    def af_empirical(self):
        """
        Empirical allele frequencies.
        Specific to HaplotypeArray objects.
        
        Returns
        -------
        numpy.ndarray
            Empirical allele frequencies.
        
        Raises
        ------
        TypeError
            If `_col_dim` is not 'variant'.
        """
        if self._col_dim != 'variant':
            raise TypeError
        hap_AF = self._obj.mean(axis=0)
        AF = np.asarray(hap_AF.data[0::2] + hap_AF.data[1::2]) / 2.
        return AF

    @property
    def maf_empirical(self):
        """
        Empirical minor allele frequencies.
        Specific to HaplotypeArray objects.
        
        Returns
        -------
        numpy.ndarray
            Empirical minor allele frequencies.
        
        Raises
        ------
        TypeError
            If `_col_dim` is not 'variant'.
        """
        if self._col_dim != 'variant':
            raise TypeError
        tmp = self.af_empirical
        tmp2 = 1 - tmp
        return np.where(tmp < tmp2, tmp, tmp2)

    @property
    def m(self):
        """
        Return the number of distinct diploid variants.
        Specific to HaplotypeArray objects.
        
        Returns
        -------
        int:
            The number of distinct diploid variants in the array.
            
        Raises
        ------
        TypeError:
            If the `_col_dim` attribute is not equal to 'variant'.
        """
        if self._col_dim != 'variant':
            raise TypeError
        return self._obj.shape[1] // 2

    def to_diploid(self):
        """
        Convert the object to a diploid representation by adding the two haplotypes for each variant.
        Specific to HaplotypeArray objects.
        
        Raises
        ------
        TypeError:
            If the `_col_dim` attribute is not equal to 'variant'.
        """
        if self._col_dim != 'variant':
            raise TypeError
        data = self._obj[:, 0::2].data + self._obj[:, 1::2].data
        vind = self.get_variant_indexer().to_diploid()
        sind = self.get_sample_indexer()

        coord_dict = sind.coord_dict.copy()
        coord_dict.update(vind.coord_dict)
        ## convert to dask array if necessary
        return xr.DataArray(data=data,
                            dims=['sample', 'variant'],
                            coords=coord_dict,
                            name='GenotypeArray',
                            attrs={
                                'generation': self._obj.generation,
                            })

        # tmp = haplo[:,0::2].data + haplo[:,1::2].data
        # ind = haplo.xft.get_variant_indexer().to_diploid()
        # xft.struct.HaplotypeArray(tmp,sample_indexer=haplo.xft.get_sample_indexer(),
        #                           variant_indexer=ind)
        raise NotImplementedError()


    def to_diploid_standardized(self, af: NDArray, scale: bool):
        """
        Standardize the HaplotypeArray object and convert it to a diploid representation.
        Specific to HaplotypeArray objects.
        
        Parameters
        ----------
        af: NDArray
            An array containing the allele frequencies of each variant.
        scale: bool
            Whether or not to scale the standardized array by the square root of the number of variants.
            
        Returns
        -------
        ndarray:
            A standardized diploid array where each variant is represented as the sum of two haplotypes.
            
        Raises
        ------
        TypeError:
            If the `_col_dim` attribute is not equal to 'variant'.
        """
        if self._col_dim != 'variant':
            raise TypeError
        if scale:
            return xft.utils.standardize_array_hw(self._obj.data, af) / np.sqrt(self.m)
        else:
            return xft.utils.standardize_array_hw(self._obj.data, af)

    def get_annotation_dict(self):
        """
        Return a dictionary of all annotation variables associated with the variants in the object.
        Specific to HaplotypeArray objects.
        
        Returns
        -------
        dict:
            A dictionary where the keys are the annotation variable names and the values are the corresponding arrays.
            
        Raises
        ------
        TypeError:
            If the `_col_dim` attribute is not equal to 'variant'.
        """
        if self._col_dim != 'variant':
            raise TypeError
        return {x[0]: x[1].values for x in self._obj.coords.variables.items() if 'variant' in x[1].dims and x[0] not in self._non_annotation_vars}

    def get_non_annotation_dict(self):
        """
        Return a dictionary of all non-annotation variables associated with the variants in the object.
        Specific to HaplotypeArray objects.
        
        Returns
        -------
        dict:
            A dictionary where the keys are the non-annotation variable names and the values are the corresponding arrays.
            
        Raises
        ------
        TypeError:
            If the `_col_dim` attribute is not equal to 'variant'.
        """
        if self._col_dim != 'variant':
            raise TypeError
        return {x[0]: x[1].values for x in self._obj.coords.variables.items() if 'variant' in x[1].dims and x[0] in self._variant_vars}

    # component index properties / methods
    def grep_component_index(self, keyword: str = 'phenotype'):
        """
        Returns the index array of components whose names contain the given keyword.
        Specific to PhenotypeArray objects.

        Parameters
        ----------
        keyword : str, optional
            The keyword to search for in component names, by default 'phenotype'.

        Returns
        -------
        XftIndex
            The index of components that match the given keyword.

        Raises
        ------
        TypeError
            If the column dimension is not 'component'.
        """
        if self._col_dim != 'component':
            raise TypeError
        pheno_cols = self._obj.component_name.values[self._obj.component_name.str.contains(
            keyword).values]
        component_index = self._obj.xft.get_component_indexer()[
            dict(component_name=pheno_cols)]
        return component_index


    def get_comp_type(self, ctype='intermediate'):
        """
        Returns the index array of components with comp_type==ctype
        Specific to PhenotypeArray objects.

        Returns
        -------
        XftIndex
            The index of components that match the given keyword.

        Raises
        ------
        TypeError
            If the column dimension is not 'component'.
        """
        if self._col_dim != 'component':
            raise TypeError
        pheno_cols = self._obj.component_name.values[self._obj.comp_type.str.contains(
            ctype).values]
        component_index = self._obj.xft.get_component_indexer()[
            dict(component_name=pheno_cols)]
        return component_index

    def get_intermediate_components(self):
        """
        Returns the index array of components with comp_type=='intermediate'
        Specific to PhenotypeArray objects.

        Returns
        -------
        XftIndex
            The index of components that match the given keyword.

        Raises
        ------
        TypeError
            If the column dimension is not 'component'.
        """
        return self.get_comp_type('intermediate')

    def get_outcome_components(self):
        """
        Returns the index array of components with comp_type=='outcome'
        Specific to PhenotypeArray objects.

        Returns
        -------
        XftIndex
            The index of components that match the given keyword.

        Raises
        ------
        TypeError
            If the column dimension is not 'component'.
        """
        return self.get_comp_type('outcome')


    @property
    def k_total(self):
        """
        Returns the total number of components.
        Specific to PhenotypeArray objects.

        Returns
        -------
        int
            The total number of components.

        Raises
        ------
        TypeError
            If the column dimension is not 'component'.
        """
        if self._col_dim != 'component':
            raise TypeError
        return self.shape[1]  # number of all phenotype components

    @property
    def k_phenotypes(self):
        """
        Returns the number of unique phenotype components.
        Specific to PhenotypeArray objects.

        Returns
        -------
        int
            The number of unique phenotype components.

        Raises
        ------
        TypeError
            If the column dimension is not 'component'.
        """
        if self._col_dim != 'component':
            raise TypeError
        return np.unique(self._obj.phenotype_name).shape[0]

    @property
    def all_phenotypes(self) -> np.ndarray:
        """
        Returns an array of all the unique phenotype component names.
        Specific to PhenotypeArray objects.

        Returns
        -------
        numpy.ndarray
            An array of all the unique phenotype component names.

        Raises
        ------
        TypeError
            If the column dimension is not 'component'.
        """

    @property
    def k_components(self) -> int:
        """
        Returns the number of unique component names.
        Specific to PhenotypeArray objects.

        Returns
        -------
        int
            The number of unique component names.

        Raises
        ------
        TypeError
            If the column dimension is not 'component'.
        """
        if self._col_dim != 'component':
            raise TypeError
        return np.unique(self._obj.component_name).shape[0]

    @property
    def all_components(self) -> np.ndarray:
        """
        Returns an array of all the unique component names.
        Specific to PhenotypeArray objects.

        Returns
        -------
        numpy.ndarray
            An array of all the unique component names.

        Raises
        ------
        TypeError
            If the column dimension is not 'component'.
        """
        if self._col_dim != 'component':
            raise TypeError
        return np.unique(self._obj.component_name)

    @property
    def k_relative(self) -> int:
        """
        Returns the number of unique origin relative values.
        Specific to PhenotypeArray objects.
        
        Returns
        -------
        int
            The number of unique origin relative values.
        
        Raises
        ------
        TypeError
            If the column dimension is not 'component'.
        """
        if self._col_dim != 'component':
            raise TypeError
        return np.unique(self._obj.vorigin_relative).shape[0]

    @property
    def all_relatives(self) -> np.ndarray:
        """
        Returns an array of all the unique origin relative values.
        Specific to PhenotypeArray objects.

        Returns
        -------
        numpy.ndarray
            An array of all the unique origin relative values.

        Raises
        ------
        TypeError
            If the column dimension is not 'component'.
        """
        if self._col_dim != 'component':
            raise TypeError
        return np.unique(self._obj.vorigin_relative)

    @property
    def k_current(self) -> int:
        """Returns the number of all current-gen specific components.
        Specific to PhenotypeArray objects.
        
        Raises
        ------
        TypeError
            If the column dimension is not 'component'.
        
        Returns
        -------
        int
            The number of all current-gen specific components.
        """
        if self._col_dim != 'component':
            raise TypeError
        return np.sum(self._obj.vorigin_relative == -1)

    def get_k_rel(self, rel) -> int:
        """Returns the number of components with the given relative origin.
        Specific to PhenotypeArray objects.
        
        Args:
            rel (int): The relative origin of the components.
        
        Raises:
            TypeError: If the column dimension is not 'component'.
            
        Returns:
            int: The number of components with the given relative origin.
        """
        if self._col_dim != 'component':
            raise TypeError
        return np.sum(self._obj.vorigin_relative == rel)

    @property
    def depth(self):
        """Returns the generational depth from binary relative encoding.
        Specific to PhenotypeArray objects.
        
        Raises
        ------
        TypeError
            If the column dimension is not 'component'.
        
        Returns
        -------
        Union[float, np.nan]
            The generational depth from binary relative encoding, or NaN if the relative origin is empty.
        """
        if self._col_dim != 'component':
            raise TypeError
        if len(self.vorigin_relative) != 0:
            return math.floor(math.log2(np.max(self._obj.vorigin_relative) + 2)) + 1
        else:
            return np.NaN

    def split_by_phenotype(self) -> Dict[str, pd.DataFrame]:
        """Splits the data by phenotype name.
        Specific to PhenotypeArray objects.
        
        Raises
        ------
        TypeError
            If the column dimension is not 'component'.
        
        Returns
        -------
        Dict[str, pd.DataFrame]
            A dictionary of dataframes, where the keys are the unique phenotype names and the values are dataframes containing the data for each phenotype.
        """
        if self._col_dim != 'component':
            raise TypeError
        return {phenotype: pheno.loc[:, pheno.phenotype_name == phenotype] for phenotype in self.all_phenotypes}

    def split_by_component(self):
        """Splits the data by component name.
        Specific to PhenotypeArray objects.
        
        Raises
        ------
        TypeError
            If the column dimension is not 'component'.
        
        Returns
        -------
        Dict[str, pd.DataFrame]
            A dictionary of dataframes, where the keys are the unique component names and the values are dataframes containing the data for each component.
        """
        if self._col_dim != 'component':
            raise TypeError
        return {component: pheno.loc[:, pheno.component_name == component] for component in self.all_components}

    def split_by_vorigin(self) -> Dict[int, pd.DataFrame]:
        """Splits the data by relative origin.
        Specific to PhenotypeArray objects.
        
        Raises
        ------
        TypeError
            If the column dimension is not 'component'.
        
        Returns
        -------
        Dict[int, pd.DataFrame]
            A dictionary of dataframes, where the keys are the unique relative origins and the values are dataframes containing the data for each relative origin.
        """
        if self._col_dim != 'component':
            raise TypeError
        return {vorigin: pheno.loc[:, pheno.vorigin_relative == vorigin] for vorigin in self.all_relatives}

    def split_by_phenotype_vorigin(self) -> Dict[Tuple[str, int], pd.DataFrame]:
        """Splits the data by phenotype name and relative origin.
        Specific to PhenotypeArray objects.
        
        Raises
        ------
        TypeError
        If the column dimension is not 'component'.
        
        Returns
        -------
        Dict[Tuple[str, int], pd.DataFrame]
            A dictionary of dataframes, where the keys are tuples of phenotype name and relative origin and the values are dataframes containing the data for each combination of phenotype name and relative origin.
        """
        if self._col_dim != 'component':
            raise TypeError
        return {(phenotype, vorigin): pheno.loc[:, (pheno.phenotype_name == phenotype) ^ (pheno.vorigin_relative == vorigin)] for phenotype in self.all_phenotypes for vorigin in self.all_relatives}

    def as_pd(self, prettify: bool = True):
        """Returns the data as a Pandas DataFrame.
        Specific to PhenotypeArray objects.
        
        Parameters
        ----------
        prettify : bool, optional
            If True, the multi-index columns will be prettified by replacing -1, 0, 1 with 'proband', 'mother', 'father', respectively.
        
        Raises
        ------
        TypeError
            If the column dimension is not 'component'.
        
        Returns
        -------
        pd.DataFrame
            A Pandas DataFrame representing the data.
        """
        if self._col_dim != 'component':
            raise TypeError
        component_mind = self.component_mindex
        if prettify:
            fr = component_mind.to_frame()
            fr['vorigin_relative'].replace(
                [-1, 0, 1], ['proband', 'mother', 'father'], inplace=True)
            component_mind = pd.MultiIndex.from_frame(fr)
        return xr.DataArray(self._obj.data,
                            dims=('sample', 'component'),
                            coords={'sample': self.sample_mindex,
                                    'component': component_mind},
                            ).to_pandas()

    def __getitem__(self, *args):
        """
        Retrieve a subset of the data with the given indices.

        Parameters
        ----------
        args : tuple
            Tuple of indices to retrieve the subset of data. It can be one of the following:
            
            - A dictionary where the keys are column names and the values are the indices of the columns to retrieve.
            - Two positional arguments representing row and column indices, respectively.

        Returns
        -------
        xr.DataArray
            The subset of data corresponding to the given indices.

        Raises
        ------
        KeyError
            If any of the indices provided are invalid.

        """
        # indexing with a dict
        argv = args[0]
        # argv = args[0]
        if isinstance(argv, dict):
            row_dict = {key: value for (
                key, value) in argv.items() if key in self.row_vars}
            col_dict = {key: value for (
                key, value) in argv.items() if key in self.column_vars}
            # TODO possible gotcha with extra args
            if len(row_dict) == 0:
                row_dict = slice(None)
            if len(col_dict) == 0:
                col_dict = slice(None)

            row_indices = self.get_row_indexer()[row_dict].unique_identifier
            column_indices = self.get_column_indexer()[
                col_dict].unique_identifier
        # indexing with two postional args
        else:
            # argv = tuple(*(args[0]))
            assert len(argv) == 2, "provide 2 positional arguments"
            # row index
            if argv[0] is None or argv[0] is slice(None):
                row_indices = slice(None)
            elif not isinstance(argv[0], xft.index.SampleIndex):
                raise KeyError
            else:
                row_indices = argv[0].unique_identifier
            # col index
            isvarindex_1 = isinstance(argv[1], xft.index.HaploidVariantIndex)
            isvarindex_2 = isinstance(argv[1], xft.index.DiploidVariantIndex)
            isvarindex = (isvarindex_1 or isvarindex_2) and (
                self._col_dim == 'variant')
            iscomindex = (isinstance(
                argv[1], xft.index.ComponentIndex) and self._col_dim == 'component')
            # print(argv)
            # print(not iscomindex)
            # print(not isinstance(argv[1], xft.index.ComponentIndex))
            if argv[1] is None or argv[1] is slice(None) or argv[1] is slice(None, None, None):
                column_indices = slice(None)
            elif (not iscomindex) and (not isvarindex):
                raise KeyError
            else:
                column_indices = argv[1].unique_identifier
            # indexing by XftIndex
        return self._obj.loc[row_indices, column_indices]

    def __setitem__(self, args, data):
        """
        Set the values of a subset of the data with the given indices.

        Parameters
        ----------
        args : tuple
            Tuple of indices to set the values of the subset of data. It can be one of the following:
            
            - A dictionary where the keys are column names and the values are the indices of the columns to set the values.
            - Two positional arguments representing row and column indices, respectively.
        data : xr.DataArray
            The data to set in the specified subset of data.

        Raises
        ------
        KeyError
            If any of the indices provided are invalid.
        """
        # indexing with a dict
        argv = args  # [0]
        # argv = args[0]
        if isinstance(argv, dict):
            row_dict = {key: value for (
                key, value) in argv.items() if key in self.row_vars}
            col_dict = {key: value for (
                key, value) in argv.items() if key in self.column_vars}
            # TODO possible gotcha with extra args
            if len(row_dict) == 0:
                row_dict = slice(None)
            if len(col_dict) == 0:
                col_dict = slice(None)

            row_indices = self.get_row_indexer()[row_dict].unique_identifier
            column_indices = self.get_column_indexer()[
                col_dict].unique_identifier
        # indexing with two postional args
        else:
            # argv = tuple(*(args[0]))
            assert len(argv) == 2, "provide 2 positional arguments"
            # row index
            if argv[0] is None or argv[0] is slice(None):
                row_indices = slice(None)
            elif not isinstance(argv[0], xft.index.SampleIndex):
                raise KeyError
            else:
                row_indices = argv[0].unique_identifier
            # col index
            isvarindex_1 = isinstance(argv[1], xft.index.HaploidVariantIndex)
            isvarindex_2 = isinstance(argv[1], xft.index.DiploidVariantIndex)
            isvarindex = (isvarindex_1 or isvarindex_2) and (
                self._col_dim == 'variant')
            iscomindex = (isinstance(
                argv[1], xft.index.ComponentIndex) and self._col_dim == 'component')

            if argv[1] is None or argv[1] is slice(None) or argv[1] is slice(None, None, None):
                column_indices = slice(None)
            elif (not iscomindex) and (not isvarindex):
                raise KeyError
            else:
                column_indices = argv[1].unique_identifier
            # indexing by XftIndex
        self._obj.loc[row_indices, column_indices] = data


class HaplotypeArray:
    """
    Represents a 2D array of binary haplotypes with accompanying row and column indices. 
    Dummy class used for generation of DataArrays and static methods 
    """
    def __new__(cls,
                # n x 2m array of binary haplotypes
                haplotypes: NDArray[Shape["*, *"], Int8] = None,
                variant_indexer: xft.index.HaploidVariantIndex = None,
                sample_indexer: xft.index.SampleIndex = None,
                generation: int = 0,
                n: int = None,
                m: int = None,
                dask: bool = False,
                **kwargs,
                ) -> xr.DataArray:
        """
        Create a new instance of DataArray.

        Parameters
        ----------
        haplotypes : np.ndarray, optional
            A 2D array of binary haplotypes. If not provided, default to None.
        variant_indexer : xft.index.HaploidVariantIndex, optional
            A haploid variant indexer. If not provided, default to None.
        sample_indexer : xft.index.SampleIndex, optional
            A sample indexer. If not provided, default to None.
        generation : int, optional
            The generation number associated with the haplotypes. Default to 0.
        n : int, optional
            The number of samples. Required if `sample_indexer` is not provided.
        m : int, optional
            The number of variants. Required if `variant_indexer` is not provided.
        dask : bool, optional
            Create a Dask array?
        **kwargs :
            Additional arguments to pass to dask.array

        Returns
        -------
        xr.DataArray
            A 2D xarray DataArray with dimensions `sample` and `variant`.

        Raises
        ------
        AssertionError
            If either `variant_indexer` or `m` is not provided, or both are provided.
            If either `sample_indexer` or `n` is not provided, or both are provided.
        
        Warns
        -----
        UserWarning
            If `haplotypes` is not provided, or if `variant_indexer` is not provided and empirical allele frequencies are used.
        """
        # obtain n,m if missing
        # assert (variant_indexer is not None) ^ (m is not None) ^ (h)), "provide variant_indexer OR m"
        # assert (sample_indexer is not None) ^ (n is not None), "provide sample_indexer OR n"
        if haplotypes is not None:
            assert haplotypes.shape[1] % 2 == 0
            n, m = haplotypes.shape
            m = m // 2
        # populate variant_indexer if missing
        if variant_indexer is None:
            variant_indexer = xft.index.DiploidVariantIndex(
                m=m, n_chrom=np.min([22, m])).to_haploid()
            if haplotypes is not None:
                warnings.warn(
                    'Using empirical allele frequencies as variant_indexer not provided')
                tmp = np.mean(haplotypes, 0)
                variant_indexer.af = np.repeat((tmp[0::2] + tmp[1::2]) * .5, 2)

        # populate sample_indexer if missing
        if sample_indexer is None:
            sample_indexer = xft.index.SampleIndex(n=n, generation=generation)
        # populate haplotypes with NaN if not provided
        if haplotypes is None:
            warnings.warn('Defaulting allele counts to -1', stacklevel=2)
            data = np.full((sample_indexer.n, variant_indexer.m * 2),
                           fill_value=-1, dtype=np.int8)
        else:
            # assert type(haplotypes) is np.ndarray, "haplotypes must be ndarray"
            data = haplotypes.astype(np.int8)

        coord_dict = sample_indexer.coord_dict.copy()
        coord_dict.update(variant_indexer.coord_dict)
        ## convert to dask array if necessary
        if dask and not isinstance(data,da.Array):
            data = da.asarray(data)
        return xr.DataArray(data=data,
                            dims=['sample', 'variant'],
                            coords=coord_dict,
                            name='HaplotypeArray',
                            attrs={
                                'generation': generation,
                            })


class PhenotypeArray:    
    """
    An array that stores phenotypes for a set of individuals.
    Dummy class used for generation of DataArrays and static methods 

    Parameters
    ----------
    components : ndarray, optional
        n x 2m array of binary haplotypes.
    component_indexer : xft.index.ComponentIndex, optional
        Indexer for components.
    sample_indexer : xft.index.SampleIndex, optional
        Indexer for samples.
    generation : int, optional
        The generation this PhenotypeArray belongs to.
    n : int, optional
        The number of samples.
    k_total : int, optional
        The total number of components.

    Returns
    -------
    xr.DataArray
        The initialized PhenotypeArray.

    Raises
    ------
    AssertionError
        If `components` is provided, then `n` and `k_total` must not be provided.
        If `component_indexer` is provided, then `k_total` must not be provided.
        If `sample_indexer` is provided, then `n` must not be provided.
        If `components` is provided and `sample_indexer` is provided, then the shape of
        `components` must match the size of the sample dimension of `sample_indexer`.
        If `components` is provided and `component_indexer` is provided, then the shape
        of `components` must match the size of the component dimension of `component_indexer`.
        If `component_indexer` is provided, then the size of the component dimension of
        `component_indexer` must match `k_total`.
    """
    def __new__(cls,
                # n x 2m array of binary haplotypes
                components: NDArray[Shape["*, *"], Float] = None,
                component_indexer: xft.index.ComponentIndex = None,
                sample_indexer: xft.index.SampleIndex = None,
                generation: int = 0,
                n: int = None,
                k_total: int = None,
                ):
        # ensure components is conformable with indexers
        if components is not None:
            assert n is None, "Provide n OR components"
            assert k_total is None, "Provide k_total OR components"
            # todo verify this doesn't induce copy
            components = np.array(components)
            if sample_indexer is not None:
                assert components.shape[0] == sample_indexer.n, "Noncomformable sample_indexer"
            if component_indexer is not None:
                assert components.shape[1] == component_indexer.k_total, "Noncomformable component_indexer"
        # obtain dimensions if necessary
        if k_total is not None:
            assert component_indexer is None, "Provide k_total OR component_indexer"
            component_indexer = xft.index.ComponentIndex(k_total=k_total)
        if n is not None:
            assert sample_indexer is None, "Provide n OR sample_indexer"
            sample_indexer = xft.index.SampleIndex(n=n)
        k_total, n = component_indexer.k_total, sample_indexer.n
        # initialize component array if necessary
        if components is None:
            components = np.full((n, k_total), fill_value=np.NaN)

        coord_dict = sample_indexer.coord_dict.copy()
        coord_dict.update(component_indexer.coord_dict)
        return xr.DataArray(data=components,
                            dims=['sample', 'component'],
                            coords=coord_dict,
                            name='PhenotypeArray',
                            attrs={
                                'generation': generation,
                            })

    @staticmethod
    def from_product(
        phenotype_name: Iterable,
        component_name: Iterable,
        vorigin_relative: Iterable,
        components: xr.DataArray = None,
        sample_indexer: xft.index.SampleIndex = None,
        generation: int = None,
        haplotypes: xr.DataArray = None,
        n: int = None,
    ) -> xr.DataArray:
        """
        Create a PhenotypeArray from a product of names.

        Parameters
        ----------
        phenotype_name : iterable
            The names of the phenotypes.
        component_name : iterable
            The names of the components.
        vorigin_relative : iterable
            The relative origins of each component.
        components : xr.DataArray, optional
            The array to use as the components.
        sample_indexer : xft.index.SampleIndex, optional
            The sample indexer to use.
        generation : int, optional
            The generation of the PhenotypeArray.
        haplotypes : xr.DataArray, optional
            The haplotypes to use.
        n : int, optional
            The number of samples to use.

        Returns
        -------
        xr.DataArray
            The new PhenotypeArray.

        Raises
        ------
        AssertionError
            If exactly one of `generation` and `sample_indexer` is provided, or exactly one
            of `haplotypes` and `sample_indexer`/`generation` or `n`/`generation` is provided.
        """
        # use either haplotypes xOR sample_indexer/generation xOR n/generation
        bool_gsi = bool(generation is not None and sample_indexer is not None)
        bool_h = bool(haplotypes is not None)
        bool_n = bool(n is not None and generation is not None)
        assert bool_gsi ^ bool_h ^ bool_n
        if bool_n:
            sample_indexer = xft.index.SampleIndex(n=n, generation=generation)
        elif bool_h:
            generation = haplotypes.xft.generation
            sample_indexer = haplotypes.xft.get_sample_indexer()
        component_indexer = xft.index.ComponentIndex.from_product(
            phenotype_name, component_name, vorigin_relative)
        return PhenotypeArray(
            components=components,
            component_indexer=component_indexer,
            sample_indexer=sample_indexer,
            generation=generation,
        )

    @staticmethod
    def _test():
        generation = 0
        n = 3
        m = 10
        n_chrom = 10
        haplotypes = np.full((n, m * 2), fill_value=-1, dtype=np.int8)
        variant_indexer = xft.index.DiploidVariantIndex(
            m=m, n_chrom=n_chrom).to_haploid()
        sample_indexer = xft.index.SampleIndex(n=n, generation=generation)
        HaplotypeArray(haplotypes, generation=generation)



import numpy as np

class Config:
    """
    A class to store configuration settings. Instantiated as xftsim.config when package is loaded

    Attributes
    ----------
    nthreads : int
        Number of threads to use for parallel execution.
    print_level : int
        Verbosity level for print statements.
    print_durations_threshold : float
        Threshold for printing durations.
    """

    def __init__(self):
        """
        Initialize the Config object with default settings.
        """
        self.nthreads = 4
        self.print_level = 2
        self.print_durations_threshold = 0. #np.inf

    def get_pdurations(self):
        """
        Get the current print durations threshold.

        Returns
        -------
        float
            The print durations threshold.
        """
        return self.print_durations_threshold

    def get_plevel(self):
        """
        Get the current print level.

        Returns
        -------
        int
            The print level.
        """
        return self.print_level


config = Config()

from . import utils       ## utility functions   
from . import data        ## download recombination maps etc
from . import index       ## indexing
from . import struct      ## data structures
from . import effect      ## genetic effects
from . import arch        ## phenogenetic architectures
from . import mate        ## mate assignment
from . import reproduce   ## sexual reproduction and phenotypic transmission
from . import founders    ## creation / import of founder haplotypes
from . import sim         ## simulation object
from . import stats       ## estimation
from . import proc        ## post-processing
from . import io          ## input/output




import pandas as pd
import pkg_resources


def get_ceu_map():
    """
    Load the CEU haplotype map.

    Returns:
    --------
    pandas.DataFrame
        A DataFrame with the CEU haplotype map.

    """
    stream = pkg_resources.resource_stream(__name__, 'maps/ceu.hg19.map')

    return pd.read_csv(stream)



"""Summary
"""
import sys
import warnings
import numpy as np
import pandas as pd
import nptyping as npt
import xarray as xr
from nptyping import NDArray, Int8, Int64, Float64, Bool, Shape
from typing import Any, Hashable, List, Iterable, Callable, Union, Dict
from functools import cached_property
from dataclasses import dataclass, field

import xftsim as xft

# simple additive effects that are standardized and/or scaled


class AdditiveEffects:
    """Additive genetic effects object. Given matrix / vector of effects will provide various scalings / offsets for computation
    
    Parameters
    ----------
    beta : NDArray[Any, Any]
        Vector of diploid effects
    variant_indexer : xft.index.HaploidVariantIndex | xft.index.DiploidVariantIndex
        Variant indexer, will determine ploidy automatically
    component_indexer : xft.index.ComponentIndex, optional
        Phenotype component indexer, defaults to xft.index.ComponentIndex.RangeIndex if not provided
    standardized : bool, optional
        True implies these are effects of standardized variants, by default True
    scaled : bool, optional
        True implies these are effects of variants * sqrt(m_causal), by default True
    
    Attributes
    ----------
    AF : NDArray
        diploid allele frequencies
    beta_scaled_standardized_diploid : NDArray
        Diploid effects scaled of standardized variants multiplied by number of causal variants per phenotype
    beta_scaled_standardized_haploid : NDArray
        haploid variant of above
    beta_scaled_unstandardized_diploid : NDArray
        Diploid effects scaled of unstandardized variants multiplied by number of causal variants per phenotype
    beta_scaled_unstandardized_haploid : NDArray
        haploid variant of above
    beta_unscaled_standardized_diploid : NDArray
        Diploid effects scaled of standardized variants unscaled by number of causal variants per phenotype
    beta_unscaled_standardized_haploid : NDArray
        haploid variant of above
    beta_unscaled_unstandardized_diploid : NDArray
        Diploid effects scaled of unstandardized variants unscaled by number of causal variants
        Multiply these against (0,1,2) raw genotypes and subtract offset to obtain phenotypes
    beta_unscaled_unstandardized_haploid : NDArray
        Haploid variant of above
    beta_raw_diploid : NDArray
        Alias for beta_unscaled_unstandardized_diploid
    beta_raw_haploid : NDArray
        Alias for beta_unscaled_unstandardized_haploid
    component_indexer : xft.index.ComponentIndex
    k : int
        Number of phenotypes (columns of effect matrix)
    m : int
        Number of diploid variants
    offset : NDArray
        To compute phenotypes, add offset after multiplying by beta_raw_* to mean deviate under HWE
    variant_indexer : xft.index.HaploidVariantIndex
    """

    def __init__(self,
        beta: NDArray[Any, Any],
        variant_indexer: Union[xft.index.HaploidVariantIndex,
            xft.index.DiploidVariantIndex] = None,
        component_indexer: xft.index.ComponentIndex = None,
        standardized: bool = True,
        scaled: bool = True,
        ):
        unscaled = not scaled
        unstandardized = not standardized
        if isinstance(variant_indexer, xft.index.HaploidVariantIndex):
            self.component_indexer = component_indexer
        else:
            self.component_indexer = component_indexer.to_haploid()

        self.variant_indexer = variant_indexer

        self._tiledAF = xft.utils.ensure2D(self.variant_indexer.af)
        self.AF = xft.utils.ensure2D(self.variant_indexer.af)[::2, :]
        if np.any(np.isnan(self.AF)):
            raise RuntimeError('Must provided allele frequencies')

        # ensure shape
        beta = xft.utils.ensure2D(beta)
        m, k = beta.shape
        self.m, self.k = variant_indexer.m, component_indexer.k_phenotypes
        if (self.m != m) or (self.k != k):
            raise RuntimeError('beta and indexers nonconformable')
        self._beta = beta

        if unscaled and standardized:
            self.beta_unscaled_unstandardized_diploid = beta / np.sqrt(2 * self.AF * (1 - self.AF))
            self.beta_unscaled_standardized_diploid = beta 
            self.beta_scaled_unstandardized_diploid = beta / np.sqrt(2 * self.AF * (1 - self.AF)) * np.sqrt(self.m_causal)
            self.beta_scaled_standardized_diploid = beta * np.sqrt(self.m_causal)
        elif scaled and unstandardized:
            self.beta_unscaled_unstandardized_diploid = beta / np.sqrt(self.m_causal)
            self.beta_unscaled_standardized_diploid = beta * np.sqrt(2 * self.AF * (1 - self.AF)) / np.sqrt(self.m_causal)
            self.beta_scaled_unstandardized_diploid = beta
            self.beta_scaled_standardized_diploid = beta * np.sqrt(2 * self.AF * (1 - self.AF))
        elif scaled and standardized:
            self.beta_unscaled_unstandardized_diploid = beta / np.sqrt(2 * self.AF * (1 - self.AF)) / np.sqrt(self.m_causal)
            self.beta_unscaled_standardized_diploid = beta / np.sqrt(self.m_causal)
            self.beta_scaled_unstandardized_diploid = beta / np.sqrt(2 * self.AF * (1 - self.AF))
            self.beta_scaled_standardized_diploid = beta
        elif unscaled and unstandardized:
            self.beta_unscaled_unstandardized_diploid = beta
            self.beta_unscaled_standardized_diploid = beta * np.sqrt(2 * self.AF * (1 - self.AF))
            self.beta_scaled_unstandardized_diploid = beta * np.sqrt(self.m_causal)
            self.beta_scaled_standardized_diploid = beta * np.sqrt(2 * self.AF * (1 - self.AF)) * np.sqrt(self.m_causal)
        else:
            raise TypeError("standardized, scaled must be bool")

        if component_indexer is None:
            self.component_indexer = xft.index.ComponentIndex.range_index(c)

    @property
    def m_causal(self) -> NDArray:  # number of non-zero betas per phenotype
        return (np.sum(self._beta != 0, axis=0))
    @property
    def offset(self) -> NDArray:  # add to raw multiple against raw (0,1,2)
        return np.sum(-2 * self.AF * self.beta_raw_diploid, axis=0)
    @property
    def beta_unscaled_unstandardized_haploid(self):
        return np.repeat(self.beta_unscaled_unstandardized_diploid, 2, axis=0)
    @property
    def beta_unscaled_standardized_haploid(self):
        return np.repeat(self.beta_unscaled_standardized_diploid, 2, axis=0)
    @property
    def beta_scaled_unstandardized_haploid(self):
        return np.repeat(self.beta_scaled_unstandardized_diploid, 2, axis=0)
    @property
    def beta_scaled_standardized_haploid(self):
        return np.repeat(self.beta_scaled_standardized_diploid, 2, axis=0)
    @property
    def beta_raw_diploid(self):
        return self.beta_unscaled_unstandardized_diploid
    @property
    def beta_raw_haploid(self):
        return self.beta_unscaled_unstandardized_haploid

    def corr(self):
        rr = np.corrcoef(self.beta_scaled_standardized_diploid, rowvar=False)
        names = self.component_indexer.phenotype_name.values
        return pd.DataFrame(rr, index=names, columns=names)


class GCTAEffects(AdditiveEffects):
    """Additive genetic effects object under GCTA infinitessimal model <CITE>

    Under this genetic architecture, all variants are causal and standardized genetic variants / sqrt(m) 
    have the user specified (possibly diagonal) variance covariance matrix

    
    Parameters
    ----------
    vg : Iterable | NDArray
        Vector of genetic variances or genetic variance/covariance matrix
    variant_indexer : xft.index.HaploidVariantIndex | xft.index.DiploidVariantIndex
        Variant indexer, will determine ploidy automatically
    component_indexer : xft.index.ComponentIndex, optional
        Phenotype component indexer, defaults to xft.index.ComponentIndex.RangeIndex if not provided
    """
    def __init__(self,
        vg: Union[Iterable, NDArray],  # either genetic variances or genetic variance/covariance matrix
        variant_indexer: Union[xft.index.HaploidVariantIndex,
            xft.index.DiploidVariantIndex] = None,
        component_indexer: xft.index.ComponentIndex = None,
        ):

        vg=np.array(vg)
        
        if len(vg.shape) == 1:
            vg=np.diag(vg)
        elif len(vg.shape) != 2:
            raise TypeError("vg must be a vector or matrix")
      
        m=variant_indexer.m
        k=component_indexer.k_phenotypes
        beta=np.random.multivariate_normal(np.zeros(k), vg, m)

        super().__init__(beta=beta,
                         variant_indexer=variant_indexer,
                         component_indexer=component_indexer,
                         scaled=True,
                         standardized=True)



class NonOverlappingEffects(AdditiveEffects):
    """Additive genetic effects object under non-infinitessimal model with no pleoitropy

    Under this genetic architecture, the genome is partitioned into k+1 components corresponding
    to k sets of variants corresponding to those causal for each trait together with a final
    set of variants not causal for any traits. Within each kth set of causal variants, 
    standardized variants are Gaussian with variance vg[k] / sqrt(proportions[k])
    
    Parameters
    ----------
    vg : Iterable
        Vector of genetic variances or genetic variance/covariance matrix
    proportions : Iterable
        Proportion of variants causal for each trait. If an extra value is provided, this will
        be the number of variants that are noncausal for all traits. Defaults to an equal
        number of variants per trait
    permute : bool
        Permute variants? If False, causal variants for each phenotype will fall into contiguous 
        blocks, defaults to True
    variant_indexer : xft.index.HaploidVariantIndex | xft.index.DiploidVariantIndex
        Variant indexer, will determine ploidy automatically
    component_indexer : xft.index.ComponentIndex, optional
        Phenotype component indexer, defaults to xft.index.ComponentIndex.RangeIndex if not provided
    """
    def __init__(self,
        vg: Iterable,
        proportions: Iterable = None,
        variant_indexer: Union[xft.index.HaploidVariantIndex,
            xft.index.DiploidVariantIndex] = None,
        component_indexer: xft.index.ComponentIndex = None,
        permute: bool = True,
        ):
     
        vg=np.array(vg).ravel()
   
        m=variant_indexer.m
        k=component_indexer.k_phenotypes   
        
        if len(vg) != k:
            raise ValueError("vg must have the same number of elements as there are phenotypes")
        
        if proportions is None:
            proportions = np.ones(k)
        elif not (len(proportions) in (k, k + 1)):
            raise ValueError("proportions must have the same number of elements as there are phenotypes or one more than that quantity")

        if permute:
            causal_inds = np.random.permutation(m)
        else:
            causal_inds = np.arange(m)

        m_causal = np.floor(xft.utils.to_simplex(proportions)*m).astype(int)
        ## if numbers per variant don't added up round by adding to biggest block
        if np.sum(m_causal) < m:
            max_ind = np.where(m_causal==np.max(m_causal))[0][0]
            m_causal[max_ind] += int(m - (np.sum(m_causal)))
        start_inds = np.cumsum(np.concatenate([[0], m_causal[:-1]]))
        stop_inds = np.cumsum(m_causal)

        beta = np.zeros((m,k))
        for j in range(k):
            beta[causal_inds[start_inds[j]:stop_inds[j]],j] = np.random.randn(m_causal[j]) * np.sqrt(vg[j])

        super().__init__(beta=beta,
                         variant_indexer=variant_indexer,
                         component_indexer=component_indexer,
                         scaled=True,
                         standardized=True)


from setuptools import setup, find_packages

setup(
    name='xftsim',
    
    version='0.1.2',
    
    author="Richard Border",
    
    author_email="border.richard@gmail.com",
    
    packages=find_packages(include=[
        'xftsim', 'xftsim*',
    ]),
    
    install_requires = [
    "funcy",
    "networkx",
    "nptyping",
    "numba",
    "numpy",
    "pandas",
    "pandas_plink",
    "scipy",
    "sgkit",
    "xarray",
    ],

    include_package_data=True,
)



import warnings
import numpy as np
import pandas as pd
import dask.array as da
import nptyping as npt
import xarray as xr
from nptyping import NDArray, Int8, Int64, Float64, Bool, Shape
from typing import Any, Hashable, List, Iterable, Union
from functools import cached_property
import numba as nb
import math
import nptyping as npt

import xftsim as xft

from xftsim.mate import MateAssignment

# todo -- base on DiploidVariantIndex


class RecombinationMap:  # diploid recombination map
    """
    A class to represent a diploid recombination map.
    In the future, will require XftIndex object instead of vid and chrom.


    Parameters
    ----------
    p : float or numpy.ndarray, optional
        Probabilities, either a float or a numpy.ndarray, default is None. A single value
        results in an exchangle map, an array corresponds to probabilities of recombination
        between specified loci
    vindex: xft.index.HaploidVariantIndex | xft.index.DiploidVariantIndex
        Variant index. Only provide if not providing vid / chrom
    vid : NDArray[Shape["*"], Any], optional
        Variant IDs, default is None.
    chrom : NDArray[Shape["*"], Int64], optional
        Chromosomes, default is None.
    """
    def __init__(self,
                 p=None,  # probabilities numpy.ndarray or float
                 vindex: Union[xft.index.HaploidVariantIndex, xft.index.DiploidVariantIndex] = None,
                 vid: NDArray[Shape["*"], Any] = None,  # variant id,
                 chrom: NDArray[Shape["*"], Int64] = None,  # chromosome
                 ):
        if vindex is not None:
            vid = vindex.vid
            chrom = vindex.chrom
        # enforce haploid
        if len(vid) == 2 * len(np.unique(vid)):
            vid = vid[::2]
            chrom = chrom[::2]
        self.vid = vid
        self.chrom = chrom
        self.m = vid.shape[0]
        self._chrom_boundary = np.concatenate(
            [[0], 1+np.where(chrom[1:] != chrom[:-1])[0]])
        if type(p) is float:
            assert p <= 1 and p >= 0, "Provide a valid probability"
            self._probabilities = np.ones(self.m) * p
        if type(p) is np.ndarray:
            assert p.shape[0] == self.m, "p and chrom must agree in length"
            self._probabilities = p

        self._probabilities[self._chrom_boundary] = .5

        self.probabilities = xr.DataArray(
            self._probabilities,
            dims=["variant"],
            coords=[vid],
            name='probabilities',
        )

    @staticmethod
    def constant_map_from_haplotypes(haplotypes=xr.DataArray,
                                     p: np.float64 = .5,
                                     ) -> "RecombinationMap":
        """
        Create a constant recombination map from haplotypes.
        
        Parameters
        ----------
        haplotypes : xr.DataArray
            Haplotypes data array.
        p : np.float64, optional
            Probability, default is 0.5.
        
        Returns
        -------
        RecombinationMap
            A constant recombination map.
        """
        vi = haplotypes.xft.get_variant_indexer()
        return RecombinationMap(p, vid=vi.vid, chrom=vi.chrom)

    @staticmethod
    def variable_map_from_haplotypes_with_cM(haplotypes=xr.DataArray,
                                             ) -> "RecombinationMap":
        """
        Create a variable recombination map from haplotypes with centimorgan distances.
        
        Parameters
        ----------
        haplotypes : xr.DataArray
            Haplotypes data array.
        
        Returns
        -------
        RecombinationMap
            A variable recombination map.
        
        Raises
        ------
        ValueError
            If distance in centimorgans is required and not present in the input.
        """
        vi_dip = haplotypes.xft.get_variant_indexer().to_diploid()
        if np.any(np.isnan(vi_dip.pos_cM)):
            raise ValueError("Distance in centimorgans required"
                             "Try interpolating with `xr.DataArray.xft.interpolate_cM`")
        d = np.diff(vi_dip.pos_cM)
        pp = (1 - np.exp(-2*d/100)) / 2
        pp[vi_dip.chrom[1:]!=vi_dip.chrom[:-1]] = .5 
        pp = np.concatenate([[.5], pp])
        return RecombinationMap(pp, vid=vi_dip.vid, chrom=vi_dip.chrom)

    def __repr__(self):
        df = pd.DataFrame.from_dict(dict(vid=self.vid,
                                         chrom=self.chrom,
                                         p=self._probabilities))
        return df.__repr__()

def transmit_parental_phenotypes(
    mating: MateAssignment,
    parental_phenotypes: xr.DataArray,
    offspring_phenotypes: xr.DataArray,
    control: dict = None,
    ) -> None:
    """
    Transmits parental phenotypes to offspring.
    
    Parameters
    ----------
    mating : MateAssignment
        An object representing mating assignments.
    parental_phenotypes : xr.DataArray
        A data array containing parental phenotypes.
    offspring_phenotypes : xr.DataArray
        A data array containing offspring phenotypes.
    control : dict, optional
        A dictionary containing additional control parameters, default is None.

    Returns
    -------
    None
    """
    # sample indexes (wrt to previous generation) for parents
    # of current generation
    parent_gen_mat_sample_ind = mating.reproducing_maternal_index
    parent_gen_pat_sample_ind = mating.reproducing_paternal_index
    # sample indexes in current generation:
    offspring_gen_mat_sample_ind = None
    offspring_gen_pat_sample_ind = None

    # component index of current generation
    offspring_component_index = parental_phenotypes.xft.get_component_indexer()
    # component indexes (in current generation) for inherited phenotypes
    offspring_gen_maternal_component_ind = offspring_component_index[dict(
        vorigin_relative=0)]
    offspring_gen_paternal_component_ind = offspring_component_index[dict(
        vorigin_relative=1)]
    # component indexes (in previous generation) for inherited phenotypes
    parent_gen_maternal_component_ind = offspring_gen_maternal_component_ind.to_proband()
    parent_gen_paternal_component_ind = offspring_gen_paternal_component_ind.to_proband()

    # transmit maternal components
    maternal_data = parental_phenotypes.xft[parent_gen_mat_sample_ind,
                                            parent_gen_maternal_component_ind].data
    offspring_phenotypes.xft[offspring_gen_mat_sample_ind,
                             offspring_gen_maternal_component_ind] = maternal_data

    # transmit paternal components
    paternal_data = parental_phenotypes.xft[parent_gen_pat_sample_ind,
                                            parent_gen_paternal_component_ind].data
    offspring_phenotypes.xft[offspring_gen_pat_sample_ind,
                             offspring_gen_paternal_component_ind] = paternal_data

    # paternal_data = parental_phenotypes.xft.access_phenotype_components(component_indexer = parent_gen_ppaternal_component_ind,
    #                                                                     sample_indexer = parent_gen_pat_sample_ind)
    # offspring_phenotypes.xft.assign_phenotype_components(data = paternal_data.data,
    #                                                      component_indexer = offspring_gen_paternal_component_ind,
    #                                                      sample_indexer = offspring_gen_pat_sample_ind)

    # offspring_gen_maternal_component_in[d
    # ## no-op if no inheritance
    # if np.all(offspring_phenotypes.vorigin_relative.values==-1):
    #     pass
    # ## else transmit phenotypes
    # else:
    #     ## columns of parent components of current generation in offspring phenotype array
    #     offspring_gen_mat_component = offspring_phenotypes[:,
    #         offspring_phenotypes.vorigin_relative==0].xft.get_component_indexer()
    #     offspring_gen_mat_uid = offspring_gen_mat_component.unique_identifier
    #     offspring_gen_pat_component = offspring_phenotypes[:,
    #         offspring_phenotypes.vorigin_relative==1].xft.get_component_indexer()
    #     offspring_gen_pat_uid = offspring_gen_pat_component.unique_identifier

    #     ## columns of parent components of current generation in parent phenotype array
    #     parent_gen_mat_component_frame = offspring_gen_mat_component.frame
    #     parent_gen_mat_component_frame.vorigin_relative = 0
    #     parent_gen_mat_component = xft.index.ComponentIndex.from_frame(parent_gen_mat_component_frame)
    #     parent_gen_mat_uid = parent_gen_mat_component.unique_identifier

    #     parent_gen_pat_component_frame = offspring_gen_pat_component.frame
    #     parent_gen_pat_component_frame.vorigin_relative = 1
    #     parent_gen_pat_component = xft.index.ComponentIndex.from_frame(parent_gen_pat_component_frame)
    #     parent_gen_pat_uid = parent_gen_pat_component.unique_identifier

    #     ## transmit maternal phenotypes if necessary
    #     if offspring_gen_mat_component.k_total > 0: ## TODO refactor to use assign by index
    #         # offspring_phenotypes.xft.assign_phenotype_components()
    #         #     parental_phenotypes.xft.access_phenotype_components(
    #         #                                                                     component_indexer=parent_gen_mat_component)
    #         # parental_phenotypes.xft.access_phenotype_components(sample_indexer=mating.reproducing_paternal_index,
    #         #                                                     component_indexer=parent_gen_mat_component)

    #         offspring_phenotypes.loc[:,
    #             offspring_gen_mat_uid] =  parental_phenotypes.loc[parent_gen_mat_sample_inds,
    #             parent_gen_mat_uid].data
    #     ## transmit paternal phenotypes if necessary
    #     if offspring_gen_pat_component.k_total > 0: ## TODO refactor to use assign by index
    #         offspring_phenotypes.loc[:,
    #             offspring_gen_pat_uid] =  parental_phenotypes.loc[parent_gen_pat_sample_inds,
    #             parent_gen_pat_uid].data


# maps recombination probabilities to haploid indicies
@nb.njit("int64[:](float64[:])")
def _meiosis_i(p):
    """
    Maps recombination probabilities to haploid indices.

    Parameters
    ----------
    p : numpy.ndarray[float64]
        An array of recombination probabilities.

    Returns
    -------
    numpy.ndarray[int64]
        An array of haploid indices.
    """
    m = p.shape[0]
    output = np.empty(m, dtype=np.int64)
    for j in range(m):
        output[j] = np.random.binomial(1, p[j])
    output = np.cumsum(output) % 2
    output += np.arange(0, 2 * m, 2)
    return output

# meiosis i and ii


@nb.njit("int8[:,:](int8[:,:], int8[:,:], int64, int64, float64[:], int64[:], int64[:], int64[:], int64[:])", parallel=True)
def _meiosis(parental_haplotypes,
             offspring_haplotypes,
             n,
             m_hap,
             recombination_p,
             maternal_inds,
             paternal_inds,
             m_meiosis_ii_inds,
             p_meiosis_ii_inds,
             ):
    """
    Performs meiosis I and II.

    Parameters
    ----------
    parental_haplotypes : numpy.ndarray[int8]
        An array of parental haplotypes.
    offspring_haplotypes : numpy.ndarray[int8]
        An array of offspring haplotypes.
    n : int64
        The number of offspring.
    m_hap : int64
        The number of haploid chromosomes.
    recombination_p : numpy.ndarray[float64]
        An array of recombination probabilities.
    maternal_inds : numpy.ndarray[int64]
        An array of maternal indices.
    paternal_inds : numpy.ndarray[int64]
        An array of paternal indices.
    m_meiosis_ii_inds : numpy.ndarray[int64]
        An array of maternal meiosis II indices.
    p_meiosis_ii_inds : numpy.ndarray[int64]
        An array of paternal meiosis II indices.

    Returns
    -------
    numpy.ndarray[int8]
        An array of offspring haplotypes.
    """
    for i in nb.prange(n):
        # maternal copies:
        m_meiosis_i_inds = _meiosis_i(recombination_p)
        p_meiosis_i_inds = _meiosis_i(recombination_p)
        for j in range(m_hap // 2):
            offspring_haplotypes[i, m_meiosis_ii_inds[j]
                                 ] = parental_haplotypes[maternal_inds[i], m_meiosis_i_inds[j]]
            offspring_haplotypes[i, p_meiosis_ii_inds[j]
                                 ] = parental_haplotypes[paternal_inds[i], p_meiosis_i_inds[j]]
    return offspring_haplotypes


def meiosis(parental_haplotypes: npt.NDArray[npt.Shape["*,*"], npt.Int8],
            recombination_p: npt.NDArray[npt.Shape["*,*"], npt.Float64],
            maternal_inds: npt.NDArray[npt.Shape["*,*"], npt.Int64],
            paternal_inds: npt.NDArray[npt.Shape["*,*"], npt.Int64],
            ) -> npt.NDArray[npt.Shape["*,*"], npt.Int8]:
    """
    Performs meiosis on parental haplotypes.

    Parameters
    ----------
    parental_haplotypes : numpy.ndarray[int8]
        An array of parental haplotypes.
    recombination_p : numpy.ndarray[float64]
        An array of recombination probabilities.
    maternal_inds : numpy.ndarray[int64]
        An array of maternal indices.
    paternal_inds : numpy.ndarray[int64]
        An array of paternal indices.

    Returns
    -------
    numpy.ndarray[int8]
        An array of offspring haplotypes.
    """
    assert (parental_haplotypes.shape[1] //
            2) == recombination_p.shape[0], "incompatable arg dimension"
    assert paternal_inds.shape[0] == maternal_inds.shape[0], "incompatable arg dimension"
    assert np.max(
        maternal_inds) <= parental_haplotypes.shape[0], "incompatable arg dimension"
    assert np.max(
        paternal_inds) <= parental_haplotypes.shape[0], "incompatable arg dimension"
    assert (parental_haplotypes.dtype == np.int8) & (
        recombination_p.dtype == np.float64), "type error"
    assert (maternal_inds.dtype == paternal_inds.dtype == np.int64), "type error"

    n = maternal_inds.shape[0]
    m_hap = parental_haplotypes.shape[1]
    offspring_haplotypes = np.empty((n, m_hap), dtype=np.int8)
    m_meiosis_ii_inds = np.arange(0, m_hap, 2, dtype=np.int64)
    p_meiosis_ii_inds = np.arange(1, m_hap, 2, dtype=np.int64)
    if isinstance(parental_haplotypes, da.Array):
        parental_haplotypes = parental_haplotypes.compute()
    return _meiosis(parental_haplotypes,
                    offspring_haplotypes,
                    n,
                    m_hap,
                    recombination_p,
                    maternal_inds,
                    paternal_inds,
                    m_meiosis_ii_inds,
                    p_meiosis_ii_inds,
                    )


class Meiosis:
    """
    A class representing the process of meiosis.

    Attributes
    ----------
    recombinationMap : RecombinationMap, optional
        A pre-defined recombination map.
    p : float, optional
        A probability used when generating an exchangable recombination map on the fly.

    Methods
    -------
    get_recombination_map(haplotypes):
        Returns the recombination map, either pre-defined or generated on the fly.
    reproduce(parental_haplotypes=None, mating=None, control=None):
        Returns a HaplotypeArray representing the offspring after meiosis.
    """
    def __init__(self,
                 rmap: RecombinationMap = None,
                 p: float = None):

        assert (p is not None) ^ (rmap is not None)
        self.recombinationMap = rmap
        self._p = p

    def get_recombination_map(self, haplotypes):
        """
        Get the recombination map, either pre-defined or generated on the fly.

        Parameters
        ----------
        haplotypes : xr.DataArray
            The haplotype data.

        Returns
        -------
        RecombinationMap
            The recombination map.
        """
        if self.recombinationMap is not None:
            return self.recombinationMap
        else:  # generate recombinationmap on the fly if necessary
            return RecombinationMap.constant_map_from_haplotypes(haplotypes, p=self._p)

    # TODO make sure recombination probability indices agree
    def reproduce(self,
                  parental_haplotypes: xr.DataArray = None,
                  mating: MateAssignment = None,
                  control: dict = None,
                  ):
        """
        Return a HaplotypeArray representing the offspring after meiosis.

        Parameters
        ----------
        parental_haplotypes : xr.DataArray, optional
            The parental haplotype data.
        mating : MateAssignment, optional
            The mate assignment object.
        control : dict, optional
            A dictionary containing control parameters.

        Returns
        -------
        HaplotypeArray
            The HaplotypeArray representing the offspring after meiosis.
        """
        rmap = self.get_recombination_map(parental_haplotypes)
        recombination_p = rmap.probabilities.values
        maternal_inds = xft.utils.match(mating.reproducing_maternal_index.unique_identifier,
                                        parental_haplotypes.sample.values)
        paternal_inds = xft.utils.match(mating.reproducing_paternal_index.unique_identifier,
                                        parental_haplotypes.sample.values)
        return xft.struct.HaplotypeArray(meiosis(parental_haplotypes.data,
                                                 recombination_p,
                                                 maternal_inds,
                                                 paternal_inds,
                                                 ),
                                         variant_indexer=parental_haplotypes.xft.get_variant_indexer(),
                                         sample_indexer=mating.offspring_sample_index,
                                         generation=parental_haplotypes.attrs['generation'] + 1)


# parental_haplotypes
# recombination_p
# maternal_inds
# paternal_inds



import xftsim as xft


import warnings
import functools
import numpy as np
import pandas as pd
import nptyping as npt
import xarray as xr
from nptyping import NDArray, Int8, Int64, Float64, Bool, Shape
from typing import Any, Hashable, List, Iterable, Callable, Union, Dict, final
from numpy.typing import ArrayLike
from functools import cached_property
import funcy



class Simulation():
    """
    A class for running an xft simulation.

    Attributes
    ----------
    mating_regime : xft.mate.MatingRegime
        Mating regime.
    recombination_map : xft.reproduce.RecombinationMap
        Recombination map.
    architecture : xft.arch.Architecture
        Phenogenetic architecture.
    statistics : Iterable, optional
        Iterable of statistics to compute each generation, by default empty list.
    post_processors : Iterable, optional
        Iterable of post processors to apply each generation, by default empty list.
    generation : int, optional
        Initial generation, by default -1, corresponding to an uninitialized simulation
    control : Dict, optional
        Control parameters for the simulation, by default an empty dictionary.
    reproduction_method : xft.reproduce.ReproductionMethod, optional
        Reproduction method for the simulation, by default xft.reproduce.Meiosis.
    control : dict
        Control parameters for the simulation
    haplotypes : xr.DataArray
        Haplotypes for the current generation.
    phenotypes : xr.DataArray
        Phenotypes for the current generation.
    mating : xr.DataArray
        Mating information for the current generation.
    parent_mating : xr.DataArray
        Mating information for the previous generation.
    parent_haplotypes : xr.DataArray
        Haplotypes for the previous generation.
    parent_phenotypes : xr.DataArray
        Phenotypes for the previous generation.
    results : xr.DataArray
        Results for the current generation.
    current_afs_empirical : xr.DataArray
        Current empirical allele frequencies.
    current_std_genotypes : xr.DataArray
        Current standardized genotypes.
    current_std_phenotypes : xr.DataArray
        Current standardized phenotypes.
    phenotype_store : Dict[int, xr.DataArray]
        Dictionary storing phenotypes for each generation.
    haplotype_store : Dict[int, xr.DataArray]
        Dictionary storing haplotypes for each generation.
    mating_store : Dict[int, xr.DataArray]
        Dictionary storing mating information for each generation.
    results_store : Dict[int, xr.DataArray]
        Dictionary storing results for each generation.
    pedigree : Any
        Pedigree information for the simulation (currently not implemented).


    Methods
    -------
    run(n_generations: int):
        Run the simulation for a specified number of generations.
    run_generation():
        Run a single generation of the simulation.
    compute_phenotypes():
        Compute phenotypes for the current generation.
    mate():
        Perform mating for the current generation.
    reproduce():
        Perform reproduction for the current generation.
    estimate_statistics():
        Estimate statistics for the current generation.
    process():
        Process the current generation using post-processors.
    update_pedigree():
        Update pedigree information for the current generation.
    increment_generation():
        Increment the current generation.
    move_forward(n_generations: int):
        Move the simulation forward by a specified number of generations.
    """
    def __init__(self,
                 founder_haplotypes: xr.DataArray,
                 mating_regime: xft.mate.MatingRegime,
                 recombination_map: xft.reproduce.RecombinationMap,
                 architecture: xft.arch.Architecture,
                 statistics: Iterable = [],
                 post_processors: Iterable = [],
                 generation: int = -1,
                 control={},
                 reproduction_method=xft.reproduce.Meiosis,
                 ):
        """
        Initialize a Simulation instance.

        Parameters
        ----------
        founder_haplotypes : xr.DataArray
            Haplotypes for the founder generation.
        mating_regime : xft.mate.MatingRegime
            Mating regime for the simulation.
        recombination_map : xft.reproduce.RecombinationMap
            Recombination map for the simulation.
        architecture : xft.arch.Architecture
            Architecture for the simulation.
        statistics : Iterable, optional
            Iterable of statistics to compute, by default an empty list.
        post_processors : Iterable, optional
            Iterable of post processors to apply, by default an empty list.
        generation : int, optional
            Initial generation, by default -1.
        control : Dict, optional
            Control parameters for the simulation, by default an empty dictionary.
        reproduction_method : xft.reproduce.ReproductionMethod, optional
            Reproduction method for the simulation, by default xft.reproduce.Meiosis.
        """
        # attributes
        self.mating_regime = mating_regime
        self.recombination_map = recombination_map
        self.architecture = architecture
        self.statistics = statistics
        self.post_processors = post_processors
        self.reproduction_method = reproduction_method
        self.reproduction_regime = self.reproduction_method(
            self.recombination_map)
        # default control parameters:
        ctrl = Simulation._default_control()
        ctrl.update(control)
        self.control = ctrl
        # properties
        self._generation = generation
        # initialize data stores
        # for proper initialization
        self.haplotype_store = {np.max([generation, 0]): founder_haplotypes}
        self.phenotype_store = {}
        self.mating_store = {}
        self.results_store = {}
        self.pedigree = None  # TODO

        #### generation specific cached properties ####
        # computed once per generation, deleted next generation
        # by call to increment_generation
        self._current_afs_empirical = None
        self._current_std_genotypes = None
        self._current_std_phenotypes = None

    @property
    def control(self):
        return self._control

    @control.setter
    def control(self, value):
        self._control = value
        self._validate_control()

    @staticmethod
    def _default_control():
        return dict(
            standardization='hardy_weinberg_empirical',
        )

    def _validate_control(self):
        if self._control['standardization'] not in [
            'hardy_weinberg_empirical',
            # 'hardy_weinberg_reference',
        ]:
            raise RuntimeError()

    @xft.utils.profiled(level=0)
    def run(self, n_generations: int):
        """
        Run the simulation for a specified number of generations.

        Parameters
        ----------
        n_generations : int
            Number of generations to run the simulation.
        """
        for generation in range(n_generations):
            self.run_generation()

    @xft.utils.profiled(level=0)
    def run_generation(self):
        """
        Run a single generation of the simulation.
        """
        self.increment_generation()
        self.reproduce()
        self.compute_phenotypes()
        self.mate()
        self.update_pedigree()
        self.estimate_statistics()
        self.process()

    @xft.utils.profiled()
    def compute_phenotypes(self):
        """
        Compute phenotypes for the current generation.
        """
        # generation zero
        if self.generation == 0:
            self.phenotypes = self.architecture.initialize_founder_phenotype_array(
                self.haplotypes, self.control)
            self.architecture.compute_phenotypes(
                self.haplotypes, self.phenotypes, self.control)
        elif self.generation >= 1:
            self.phenotypes = self.architecture.initialize_phenotype_array(
                self.haplotypes, self.control)
            xft.reproduce.transmit_parental_phenotypes(self.parent_mating,
                                                       self.parent_phenotypes,
                                                       self.phenotypes,
                                                       self.control)
            self.architecture.compute_phenotypes(
                self.haplotypes, self.phenotypes, self.control)

    @xft.utils.profiled()
    def mate(self):
        """
        Perform mating for the current generation.
        """
        self.mating = self.mating_regime.mate(
            self.haplotypes, self.phenotypes, self.control)

    @xft.utils.profiled()
    def reproduce(self):
        """
        Perform reproduction for the current generation.
        """
        if self.generation == 0:
            pass
        elif self.generation >= 1:
            self.haplotypes = self.reproduction_regime.reproduce(self.parent_haplotypes,
                                                                 self.parent_mating,
                                                                 self.control)

    @xft.utils.profiled()
    def estimate_statistics(self):
        """
        Estimate statistics for the current generation.
        """
        for stat in self.statistics:
            stat.estimate(self)

    @xft.utils.profiled()
    def process(self):
        """
        Apply post-processors to the current generation.
        """
        for ind, proc in enumerate(self.post_processors):
            # allow callables that map xft.Simulation -> None
            if isinstance(proc, Callable):
                proc = xft.proc.PostProcessor(
                    proc, ' '.join(['process', str(ind)]))
            proc.process(self)

    @xft.utils.profiled()
    def update_pedigree(self):
        """
        Update pedigree information (NOT IMPLEMENTED).
        """
        # warnings.warn('update_pedigree() not implemented')
        pass  # TODO

    # generation is immutable except through Simulation().increment_generation()
    @property
    def generation(self):
        return self._generation

    # erase current-generation specific cached properties
    @xft.utils.profiled()
    def increment_generation(self):
        self._current_afs_empirical = None
        self._current_std_genotypes = None
        self._current_std_haplotypes = None
        self._generation += 1

    @property
    def haplotypes(self):
        if self.generation in self.haplotype_store.keys():
            return self.haplotype_store[self.generation]
        else:
            return None

    @haplotypes.setter
    def haplotypes(self, value):
        self.haplotype_store[self.generation] = value

    @property
    def phenotypes(self):
        if self.generation in self.phenotype_store.keys():
            return self.phenotype_store[self.generation]
        else:
            return None

    @phenotypes.setter
    def phenotypes(self, value):
        self.phenotype_store[self.generation] = value

    @property
    def mating(self):
        if self.generation in self.mating_store.keys():
            return self.mating_store[self.generation]
        else:
            return None

    @mating.setter
    def mating(self, value):
        self.mating_store[self.generation] = value

    @property
    def parent_mating(self):
        if (self.generation - 1) in self.mating_store.keys():
            return self.mating_store[(self.generation - 1)]
        else:
            return None

    @property
    def parent_haplotypes(self):
        if (self.generation - 1) in self.haplotype_store.keys():
            return self.haplotype_store[(self.generation - 1)]
        else:
            return None

    @property
    def parent_phenotypes(self):
        if (self.generation - 1) in self.phenotype_store.keys():
            return self.phenotype_store[(self.generation - 1)]
        else:
            return None

    def move_forward(self, n_generations):
        pass

    @property
    def results(self):
        if self.generation in self.results_store.keys():
            return self.results_store[self.generation]
        else:
            return None

    # generation specific properties that are overwritten
    # by increment_generation()
    @property
    def current_afs_empirical(self):
        if self._current_afs_empirical is None:
            return self.haplotypes.xft.af_empirical
        else:
            return self._current_afs_empirical

    @property
    def current_std_genotypes(self):
        if self._current_std_genotypes is None:
            if self.control['standardization'] == 'hardy_weinberg_empirical':
                return self.haplotypes.xft.to_diploid_standardized(af=self.current_afs_empirical,
                                                                   scale=False)
            else:
                raise NotImplementedError()
        else:
            return self._current_std_genotypes

    @property
    def current_std_phenotypes(self):
        if self._current_std_phenotypes is None:
            return self.phenotypes.xft.standardize()
        else:
            return self._current_std_phenotypes

    # def __repr__(self):
        # pass

    @property
    def dependency_graph_edges(self):
        edges = []
        colors = []
        edge_labels = []
        for i,component in enumerate(self.architecture.components):
            edges += component.dependency_graph_edges
            colors += [i+1 for j in range(len(component.dependency_graph_edges))]
            edge_labels += [component._component_name for j in range(len(component.dependency_graph_edges))]
        if 'proband\nhaplotypes' in [edge[0] for edge in edges]:
            edges += [('maternal\nhaplotypes','proband\nhaplotypes'),
                      ('paternal\nhaplotypes','proband\nhaplotypes')]
            colors += [0, 0]
            edge_labels += ['meiosis', 'meiosis']
        ## hack to ensure noise nodes (labelled #\d+) have unique blank labels
        import re
        noise_nodes = [x[0] for x in [re.findall('^#\\d+$', node,) for node in [edge[0] for edge in edges]] if x != []]
        noise_dict = {x:' '*(i+1) for i,x in enumerate(np.unique(noise_nodes))}
        for i in range(len(edges)):
            edge = edges[i]
            if edge[0] in noise_nodes:
                edges[i] = (noise_dict[edge[0]],edge[1])
        ## add mating
        edges += self.mating_regime.dependency_graph_edges
        colors += [np.max(colors)+1 for edge in self.mating_regime.dependency_graph_edges]
        edge_labels += ['mating' for edge in self.mating_regime.dependency_graph_edges]
        return (edges,np.array(colors), edge_labels)

    @property
    def dependency_graph(self):
        import networkx as nx
        G = nx.DiGraph()
        edges, colors, edge_labels = self.dependency_graph_edges
        G.add_edges_from(edges)
        pos = nx.nx_agraph.graphviz_layout(G, prog='dot')
        return (G,pos,colors,edges,edge_labels)

    def draw_dependency_graph(self, 
                              node_color='none', 
                              node_size = 1200, 
                              font_size=5, 
                              margins=.1,
                              edge_color="#222222", 
                              arrowsize = 6,
                              number_edges: bool = True,
                              **kwargs):
        import networkx as nx
        G,pos,colors,edges,edge_labels = self.dependency_graph
        if number_edges:
            edge_labels = xft.utils.paste([colors.astype(str),edge_labels],sep=':')
        color_dict = {a:b for a,b in zip(edges, colors)}
        label_dict = {a:b for a,b in zip(edges, edge_labels)}
        new_colors = [color_dict[edge] for edge in G.edges]
        # new_labels = [label_dict[edge] for edge in G.edges]
        # labels = {key:str(value) for (key,value) in zip(G.edges, colors)}
        nx.draw_networkx(G,pos, 
                         node_color=node_color, 
                         node_size = node_size, 
                         font_size=font_size,
                         margins=margins,
                         edge_color=new_colors,
                         arrowsize=arrowsize,
                         **kwargs)
        nx.draw_networkx_edge_labels(G,pos, 
                 font_size=font_size//1.3,
                 edge_labels=label_dict,
                 rotate=False,
                 bbox={'boxstyle':'round', 'ec':'none', 'fc':(1.0, 1.0, 1.0)},
                 # font_color=colors/np.max(colors),
                 )



class DemoSimulation(Simulation):
    demo_routines = dict(UGRM = 'Univariate GCTA with balanced random mating demo\n',
                         BGRM = 'Bivariate GCTA with balanced random mating demo\n')
    _demo_long = dict(BGRM = 'Two phenotypes, height and bone mineral denisty (BMD)\n'
                             'assuming bivariate GCTA infinitessimal archtecture\n'
                             'with h2 values set to 0.5 and 0.4 for height and BMD\n'
                             'respectively and a genetic effect correlation of 0.0.',
                      UGRM = 'A single phenotype, height, assuming univariate GCTA\n'
                             'infinitessimal archtecture with h2 = 0.5.')
    def __init__(self, 
                 routine: str = 'BGRM',
                 n: int = 2000, 
                 m: int = 400) -> None:
        if routine == 'UGRM':
            founder_haplotypes = xft.founders.founder_haplotypes_uniform_AFs(n=n, m=m)
            architecture = xft.arch.GCTA_Architecture(h2=[.5], phenotype_name=['height'], 
                                                      haplotypes=founder_haplotypes)
            recombination_map = xft.reproduce.RecombinationMap.constant_map_from_haplotypes(founder_haplotypes, 
                                                                                            p =.1)
            mating_regime = xft.mate.RandomMatingRegime(mates_per_female=1,
                                                        offspring_per_pair=2)
        elif routine == 'BGRM':
            founder_haplotypes = xft.founders.founder_haplotypes_uniform_AFs(n=n, m=m)
            architecture = xft.arch.GCTA_Architecture(h2=[.5,.4], phenotype_name=['height', 'BMD'], 
                                                      haplotypes=founder_haplotypes)
            recombination_map = xft.reproduce.RecombinationMap.constant_map_from_haplotypes(founder_haplotypes, 
                                                                                            p =.1)
            mating_regime = xft.mate.RandomMatingRegime(mates_per_female=1,
                                                        offspring_per_pair=2)
        else:
            raise NotImplementedError(f"Routine `{routine}` not implemented. See `DemoSimulation.demo_routines` for supported routines")
        ## ensure silence
        old_config = xft.config
        xft.config.print_durations_threshold = np.inf
        super().__init__(founder_haplotypes=founder_haplotypes,
                         mating_regime=mating_regime,
                         recombination_map=recombination_map,
                         architecture=architecture,
                         statistics=[xft.stats.SampleStatistics(),
                            xft.stats.MatingStatistics(full=True)])
        self.run(2)
        ## revert to user specified config
        xft.config.print_durations_threshold = old_config.print_durations_threshold
        self.routine = routine
        self.routine_title = self.demo_routines[routine]
        self.routine_long = self._demo_long[routine]
        self._n = n
        self._m = m

    def __repr__(self):
        return '\n'.join(["<DemoSimulation>",
                         self.routine_title,
                         f"n = {self._n}; m = {self._m}",
                         self.routine_long,
                         ])



"""
This module contains functions and classes for implementing different mating regimes in the context of forward time genetics simulations.

Functions:

    _solve_qap_ls: Private function that solves the Quadratic Assignment Problem using LocalSolver.

Classes:

    MatingRegime: Base class for defining mating regimes.
    RandomMatingRegime: A class for implementing random mating.
    LinearAssortativeMatingRegime: A class for implementing linear assortative mating.
    KAssortativeMatingRegime: A class for implementing k-assortative mating.
    BatchedMatingRegime: A class for batching individuals to improve mating regime performance.
"""
import warnings
import numpy as np
import pandas as pd
import nptyping as npt
import xarray as xr
from nptyping import NDArray, Int8, Int64, Float64, Bool, Shape
from typing import Any, Hashable, List, Iterable, Callable, Union, Dict
from functools import cached_property
from dataclasses import dataclass, field
import math

import xftsim as xft


## todo add methods for offspring per iid etc
## todo add methods for checking validity
## TODO properties
class MateAssignment:
    """
    Represents a mate assignment for a given generation of individuals.

    Parameters
    ----------
    generation : int
        The generation number.
    maternal_sample_index : xft.index.SampleIndex
        The sample index for the maternal individuals.
    paternal_sample_index : xft.index.SampleIndex
        The sample index for the paternal individuals.
    previous_generation_sample_index : xft.index.SampleIndex
        The sample index for the previous generation.
    n_offspring_per_pair : NDArray[Shape["*"], Int64]
        An array containing the number of offspring per mating pair.
    n_females_per_pair : NDArray[Shape["*"], Int64]
        An array containing the number of female offspring per mating pair.
    sex_aware : bool, optional (default=False)
        Whether the mate assignment is sex-aware.
    """
    def __init__(self,
                 generation: int,
                 maternal_sample_index: xft.index.SampleIndex, 
                 paternal_sample_index: xft.index.SampleIndex, 
                 previous_generation_sample_index: xft.index.SampleIndex, 
                 n_offspring_per_pair: NDArray[Shape["*"], Int64],
                 n_females_per_pair: NDArray[Shape["*"], Int64],
                 sex_aware: bool=False,
                 ):
        ## catch duplicates ## TODO
        # maternal_iids,paternal_iids,n_offspring_per_pair = xft.utils.merge_duplicate_pairs(maternal_iids,
                                                                                  # paternal_iids,
                                                                                  # n_offspring_per_pair,
                                                                                  # sort = not sex_aware)
        self._maternal_sample_index = maternal_sample_index
        self._paternal_sample_index = paternal_sample_index
        self.generation = generation
        self.n_offspring_per_pair = np.array(n_offspring_per_pair)
        self.n_females_per_pair = np.array(n_females_per_pair)
        self.n_males_per_pair = self.n_offspring_per_pair - self.n_females_per_pair
        self._family_inds = np.arange(maternal_sample_index.n)
        self.previous_generation_sample_index = previous_generation_sample_index
        self.sex_aware = sex_aware

    def get_mate_phenotypes(self, 
                        phenotypes: xr.DataArray, 
                        component_index: xft.index.ComponentIndex = None,
                        full: bool = True):
        """
        Retrieves mate phenotypes based on the given phenotypes data.

        Parameters
        ----------
        phenotypes : xr.DataArray
            The phenotypes data array.
        component_index : xft.index.ComponentIndex, optional
            The component index for the phenotypes data array.
        full : bool
            Ignore component_index and get all components.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the mate phenotypes.
        """
        if full:
            component_index = phenotypes.xft.get_component_indexer()
        elif component_index is None:
            component_index = phenotypes.xft.grep_component_index('phenotype')[dict(vorigin_relative=-1)]
        Y_maternal = phenotypes.xft[self._maternal_sample_index, component_index]
        Y_paternal = phenotypes.xft[self._paternal_sample_index, component_index]
        new_index = xft.utils.paste([Y_maternal.iid.values,
                                     Y_paternal.iid.values], sep = ":")        
        Y_maternal = Y_maternal.xft.reindex_components(Y_maternal.xft.get_component_indexer().to_vorigin(0)).to_pandas().set_index(new_index)
        Y_paternal = Y_paternal.xft.reindex_components(Y_paternal.xft.get_component_indexer().to_vorigin(1)).to_pandas().set_index(new_index)
        return pd.concat([Y_maternal, Y_paternal], axis=1)


    @staticmethod
    def reduce_merge(assignments: Iterable):
        """
        Merges a list of MateAssignment objects into a single MateAssignment object.

        Parameters
        ----------
        assignments : Iterable
            An iterable of MateAssignment objects to be merged.

        Returns
        -------
        MateAssignment
            A new MateAssignment object resulting from the merge of the input assignments.
        """
        return MateAssignment(
                              generation = int(np.max([x.generation for x in assignments])),
                              maternal_sample_index = xft.index.XftIndex.reduce_merge([x._maternal_sample_index for x in assignments]),
                              paternal_sample_index = xft.index.XftIndex.reduce_merge([x._paternal_sample_index for x in assignments]), 
                              n_offspring_per_pair = np.concatenate([x.n_offspring_per_pair for x in assignments]),
                              n_females_per_pair = np.concatenate([x.n_females_per_pair for x in assignments]),
                              # sex_aware = bool(np.prod([x.sex_aware for x in assignments])),      
                              previous_generation_sample_index = xft.index.XftIndex.reduce_merge([x.previous_generation_sample_index for x in assignments]), 
                              )
    @property
    def _expanded_indices(self):
        """
        A NumPy array containing the expanded indices for the offspring.

        Returns
        -------
        np.ndarray
            An array containing the expanded indices for the offspring.
        """
        return np.repeat(range(self._maternal_sample_index.n),
                         self.n_offspring_per_pair)

    @property
    def reproducing_maternal_index(self):
        """
        The maternal index for reproducing individuals.

        Returns
        -------
        xft.index.SampleIndex
            The maternal index for reproducing individuals.
        """
        return self._maternal_sample_index.iloc(self._expanded_indices)
    
    @property
    def reproducing_paternal_index(self):
        """
        The paternal index for reproducing individuals.

        Returns
        -------
        xft.index.SampleIndex
            The paternal index for reproducing individuals.
        """
        return self._paternal_sample_index.iloc(self._expanded_indices)

    @property
    def n_females(self):
        """
        The total number of female offspring.

        Returns
        -------
        int
            The total number of female offspring.
        """
        return np.sum(self.n_females_per_pair)

    @property
    def n_males(self):
        """
        The total number of male offspring.

        Returns
        -------
        int
            The total number of male offspring.
        """
        return np.sum(self.n_males_per_pair)

    @property
    def n_reproducing_pairs(self):
        """
        The total number of reproducing pairs.

        Returns
        -------
        int
            The total number of reproducing pairs.
        """
        return np.sum(self.n_offspring_per_pair>=1)
    
    @property
    def n_total_offspring(self):
        """
        The total number of offspring.

        Returns
        -------
        int
            The total number of offspring.
        """
        return np.sum(self.n_offspring_per_pair)

    @property
    def offspring_iids(self):
        """
        The unique identifiers for the offspring.

        Returns
        -------
        np.ndarray
            An array containing the unique identifiers for the offspring.
        """
        return xft.utils.ids_from_generation_range(self.generation+1, self.n_total_offspring) ## TODO possible gotcha with chunking?
    @property
    def offspring_fids(self):
        """
        The family identifiers for the offspring.

        Returns
        -------
        np.ndarray
            An array containing the family identifiers for the offspring.
        """
        return xft.utils.ids_from_generation(self.generation+1, self._family_inds)[self._expanded_indices]
    @property
    def offspring_sex(self):
        """
        The sex of the offspring.

        Returns
        -------
        np.ndarray
            An array containing the sex of the offspring.
        """
        return np.concatenate([np.repeat([0,1],[x,y]) for (x,y) in zip(self.n_females_per_pair,
                                                                       self.n_males_per_pair)])

    @property
    def is_constant_population(self):
        """
        TODO property to determine if the population is constant or not.

        Returns
        -------
        bool
            True if the population is constant, False otherwise.
        """
        raise NotImplementedError ## TODO

    @cached_property## TODO possible gotcha with chunking?
    def offspring_sample_index(self):
        """
        The sample index for the offspring.

        Returns
        -------
        xft.index.SampleIndex
            The sample index for the offspring.
        """
        return xft.index.SampleIndex(iid=self.offspring_iids,
                                    fid=self.offspring_fids,
                                    sex=self.offspring_sex)


    def get_mating_frame(self):
        """
        Constructs a DataFrame containing mate phenotypes regardless of reproductive success.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing mating information.
        """
        frame = np.hstack([self._maternal_sample_index.coord_frame.to_numpy(),
                           self._paternal_sample_index.coord_frame.to_numpy(),
                           xft.utils.ensure2D(self.n_offspring_per_pair),
                           xft.utils.ensure2D(self.n_females_per_pair),
                           ])
        frame = pd.DataFrame(frame, columns=[
                                            'maternal_sample', 'maternal_iid', 'maternal_fid', 'maternal_sex',
                                            'paternal_sample', 'paternal_iid', 'paternal_fid', 'paternal_sex',
                                            'n_offspring',
                                            'n_female_spring',
                                            ])
        return frame

    def get_reproduction_frame(self):
        """
        Constructs a DataFrame containing information relating to mates and offspring.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing reproduction information.
        """
        frame = np.hstack([self.offspring_sample_index.coord_frame.to_numpy(),
                           self.reproducing_maternal_index.coord_frame.to_numpy(),
                           self.reproducing_paternal_index.coord_frame.to_numpy(),
                           ])
        frame = pd.DataFrame(frame, columns=['sample', 'iid', 'fid', 'sex',
                                            'maternal_sample', 'maternal_iid', 'maternal_fid', 'maternal_sex',
                                            'paternal_sample', 'paternal_iid', 'paternal_fid', 'paternal_sex',
                                            ])
        return frame.set_index('sample')

    @property
    def maternal_integer_index(self):
        """
        The integer index for the maternal individuals.

        Returns
        -------
        np.ndarray
            An array containing the integer index for the maternal individuals.
        """
        return xft.utils.match(self.reproducing_maternal_index.unique_identifier, 
                              self.previous_generation_sample_index.unique_identifier)
    @property
    def paternal_integer_index(self):
        """
        The integer index for the paternal individuals.

        Returns
        -------
        np.ndarray
            An array containing the integer index for the paternal individuals.
        """
        return xft.utils.match(self.reproducing_paternal_index.unique_identifier, 
                              self.previous_generation_sample_index.unique_identifier)

    def update_pedigree(self, pedigree):
        raise NotImplementedError ## TODO

    def trio_view(self, pheno_parental, pheno_offspring):
        """
        Returns an array with the phenotypes of offspring, followed by the phenotypes of their parents
        in the same order as the order of offspring in this MateAssignment.

        Parameters
        ----------
        pheno_parental : xr.DataArray
            An xarray DataArray containing the phenotypes of the parents.
        pheno_offspring : xr.DataArray
            An xarray DataArray containing the phenotypes of the offspring.

        Returns
        -------
        np.ndarray
            An array with the phenotypes of offspring, followed by the phenotypes of their parents.

        """
        return np.hstack([pheno_offspring.data, 
                          pheno_parental.data[self.maternal_integer_index],
                          pheno_parental.data[self.paternal_integer_index]])

    # @property
    # def reproducing_maternal_indexer(self):
    #     return self.all_maternal_iids[self._expanded_indices]
    # @property
    # def reproducing_paternal_indexer(self):
    #     return self.all_paternal_iids[self._expanded_indices]




# ma = MateAssignment(0,
#                     maternal_sample_index=xft.index.SampleIndex(['a','b','c','d','a','b']),
#                     paternal_sample_index=xft.index.SampleIndex(['A','B','C','D','A','B']),
#                     previous_generation_sample_index = xft.index.SampleIndex(['a','b','c','d','a','b','A','B','C','D','A','B']),
#                     n_offspring_per_pair=[1,0,2,1,2,3],
#                     n_females_per_pair=[1,0,0,1,1,2])


# ## TODO
# def mergeAssignments(assignments: Iterable = None):
#     pass




class MatingRegime:
    """
    A class for defining a mating regime to simulate the reproductive behavior of a population.

    Parameters
    ----------
    mateFunction : Callable, optional
        A function that specifies how the mating process is carried out. Default is None.
    offspring_per_pair : Union[Callable, int, xft.utils.VariableCount], optional
        The number of offspring per mating pair. This can be a callable function, an integer, or a VariableCount object. Default is xft.utils.ConstantCount(1).
    mates_per_female : Union[Callable, int, xft.utils.VariableCount], optional
        The number of mating partners each female has. This can be a callable function, an integer, or a VariableCount object. Default is xft.utils.ConstantCount(1).
    female_offspring_per_pair : Union[Callable, str, int, xft.utils.VariableCount], optional
        The number of female offspring per mating pair. This can be a callable function, a string, an integer, or a VariableCount object. If set to 'balanced', the number of female offspring will be randomly assigned from a balanced range (0, ..., total_offspring). Default is 'balanced'.
    sex_aware : bool, optional
        Whether the mating process should take sex into account. If True, females and males will be paired up based on their sex. If False, the pairs will be randomly assigned. Default is False.
    exhaustive : bool, optional
        Whether the mating pairs should be enumerated exhaustively or randomly. If True, all possible pairings will be enumerated before repeating. If False, the pairings will be randomly assigned with replacement. Default is True.
    component_index : xft.index.ComponentIndex, optional
        Which phenotype components (if any) are used in assigning mates
    haplotypes : bool, optional
        Flag indeicating if haplotype data is used to assign mates (defaults to False)

    Attributes
    ----------
    sex_aware : bool
        Whether the mating process should take sex into account.
    offspring_per_pair : Union[Callable, int, xft.utils.VariableCount]
        The number of offspring per mating pair.
    mates_per_female : Union[Callable, int, xft.utils.VariableCount]
        The number of mating partners each female has.
    female_offspring_per_pair : Union[Callable, str, int, xft.utils.VariableCount]
        The number of female offspring per mating pair.
    exhaustive : bool
        Whether the mating pairs should be enumerated exhaustively or randomly.
    mateFunction : Callable
        A function that specifies how the mating process is carried out.
    expected_offspring_per_pair : float
        The expected number of offspring per mating pair.
    expected_mates_per_female : float
        The expected number of mating partners each female has.
    expected_female_offspring_per_pair : float
        The expected number of female offspring per mating pair.
    population_growth_factor : float
        The population growth factor.

    Methods
    -------
    get_potential_mates(haplotypes: xr.DataArray = None, phenotypes: xr.DataArray = None) -> (NDArray, NDArray)
        Returns the potential female and male mating partners based on the sex awareness parameter.
    enumerate_assignment(female_indices: NDArray, male_indices: NDArray, haplotypes: xr.DataArray = None, phenotypes: xr.DataArray = None) -> MateAssignment
        Enumerates the mating assignments.
    mate(haplotypes: xr.DataArray = None, phenotypes: xr.DataArray = None, control: dict = None) -> MateAssignment
        Calls the mateFunction to perform the mating process.
    """
    def __init__(self, 
                 mateFunction: Callable = None,
                 offspring_per_pair: Union[Callable, int, xft.utils.VariableCount] = xft.utils.ConstantCount(1),
                 mates_per_female: Union[Callable, int, xft.utils.VariableCount] =  xft.utils.ConstantCount(1),
                 female_offspring_per_pair: Union[Callable, str, int, xft.utils.VariableCount] = 'balanced', ## doesn't make total sense
                 sex_aware: bool = False,
                 exhaustive: bool = True,
                 component_index: xft.index.ComponentIndex = None,
                 haplotypes: bool = False,
                 ):
        ## replace integer counts with ConstantCount objects
        if isinstance(offspring_per_pair, int):
            offspring_per_pair = xft.utils.ConstantCount(offspring_per_pair)
        if isinstance(mates_per_female, int):
            mates_per_female = xft.utils.ConstantCount(mates_per_female)
        if isinstance(female_offspring_per_pair, int):
            female_offspring_per_pair = xft.utils.ConstantCount(female_offspring_per_pair)
        ## set attributes
        self.sex_aware = sex_aware
        self.offspring_per_pair = offspring_per_pair
        self.mates_per_female = mates_per_female
        self.female_offspring_per_pair = female_offspring_per_pair
        self.exhaustive = exhaustive
        self._mateFunction = mateFunction
        self._component_index = component_index
        self._haplotypes = haplotypes

    def get_potential_mates(self,
                            haplotypes: xr.DataArray = None, 
                            phenotypes: xr.DataArray = None,
                            ) -> (NDArray, NDArray):
        """
        Return potential mating pairs.

        Parameters
        ----------
        haplotypes: xr.DataArray
            The haplotypes to use for mating.
        phenotypes: xr.DataArray
            The phenotypes to use for mating.

        Returns
        -------
        (NDArray, NDArray)
            The potential female and male mating indices.
        """
        self._sample_indexer = haplotypes.xft.get_sample_indexer()

        if self.sex_aware:
            female_indices = haplotypes.sample[haplotypes.sex==0]
            male_indices = haplotypes.sample[haplotypes.sex==1]
            assert len(female_indices) == len(male_indices), "Unbalanced population"
        elif not self.sex_aware:
            permuted_indices = np.random.permutation(haplotypes.sample)
            female_indices = np.sort(permuted_indices[0::2])
            male_indices = np.sort(permuted_indices[1::2])

        return (female_indices, male_indices)

    def enumerate_assignment(self,
                             female_indices: NDArray,
                             male_indices: NDArray,
                             haplotypes: xr.DataArray = None, 
                             phenotypes: xr.DataArray = None,
                             ) -> MateAssignment:
        """
        Enumerate the mate assignments.

        Parameters
        ----------
        female_indices: NDArray
            The indices of the females to mate.
        male_indices: NDArray
            The indices of the males to mate.
        haplotypes: xr.DataArray
            The haplotypes to use for mating.
        phenotypes: xr.DataArray
            The phenotypes to use for mating.

        Returns
        -------
        MateAssignment
            The mate assignments.
        """
        matings_per_female = self.mates_per_female.draw(len(female_indices))
        n_matings = np.sum(matings_per_female)

        offspring_per_mate_pair = self.offspring_per_pair.draw(n_matings)

        female_mate_indices = xft.utils.exhaustive_enumerate(female_indices, matings_per_female)
        if self.exhaustive:
            male_mate_indices = xft.utils.exhaustive_enumerate(male_indices, matings_per_female)
        else:
            raise NotImplementedError#male_mate_indices = np.random.choice(male_indices, n_matings)

        if not isinstance(self.female_offspring_per_pair, str):
            raise NotImplementedError
        elif self.female_offspring_per_pair == 'balanced':
                n_females_per_pair = np.apply_along_axis(lambda x: np.random.randint(x+1), 0, offspring_per_mate_pair)
        else:            
            raise NotImplementedError
        if self.exhaustive:
            male_mate_indices = xft.utils.exhaustive_enumerate(male_indices, matings_per_female)
        else:
            raise NotImplementedError
            # male_mate_indices = np.random.choice(male_indices, n_matings)

        return MateAssignment(generation = haplotypes.generation,
                              maternal_sample_index = self._sample_indexer[female_mate_indices],
                              paternal_sample_index = self._sample_indexer[male_mate_indices],
                              previous_generation_sample_index = self._sample_indexer,
                              n_offspring_per_pair = offspring_per_mate_pair,
                              n_females_per_pair = n_females_per_pair,
                              sex_aware = self.sex_aware,
                              )

    @property
    def mateFunction(self):
        if self._mateFunction is not None:
            return self._mateFunction
        else:
            raise NotImplementedError("'mateFunction' not implemented")


    @mateFunction.setter
    def mateFunction(self, value):
        self._mateFunction = value

    def mate(self,
            haplotypes: xr.DataArray = None,
            phenotypes: xr.DataArray = None,
            control: dict = None,
             ) -> MateAssignment:
        """
        Mate individuals.

        Parameters
        ----------
        haplotypes : xarray.DataArray, optional
            The haplotypes of the individuals, by default None.
        phenotypes : xarray.DataArray, optional
            The phenotypes of the individuals, by default None.
        control : dict, optional
            The mating control parameters, by default None.

        Returns
        -------
        MateAssignment
            The mate assignment result.
        """
        return self._mateFunction(haplotypes, phenotypes)

    @property
    def expected_offspring_per_pair(self) -> float:
        """
        Get the expected offspring per pair.

        Returns
        -------
        float
            The expected offspring per pair.

        Raises
        ------
        NotImplementedError
            If the offspring count is not an integer or a VariableCount.
        """
        if type(self.offspring_per_pair) is int:
            return self.offspring_per_pair
        elif isinstance(self.offspring_per_pair, xft.utils.VariableCount):
            return self.offspring_per_pair.expectation
        else:
            raise NotImplementedError
    
    @property
    def expected_mates_per_female(self) -> float:
        """
        Get the expected mates per female.

        Returns
        -------
        float
            The expected mates per female.

        Raises
        ------
        NotImplementedError
            If the mates count is not an integer or a VariableCount.
        """
        if type(self.mates_per_female) is int:
            return self.mates_per_female
        elif isinstance(self.mates_per_female, xft.utils.VariableCount):
            return self.mates_per_female.expectation
        else:
            raise NotImplementedError

    @property
    def expected_female_offspring_per_pair(self) -> float:
        """
        Get the expected female offspring per pair.

        Returns
        -------
        float
            The expected female offspring per pair.

        Raises
        ------
        NotImplementedError
            If the female offspring count is not an integer or a VariableCount.
        """
        if type(self.female_offspring_per_pair) is int:
            return self.female_offspring_per_pair
        elif isinstance(self.female_offspring_per_pair, xft.utils.VariableCount):
            return self.female_offspring_per_pair.expectation
        else:
            raise NotImplementedError

    @property
    def population_growth_factor(self) -> float:
        """
        Get the population growth factor.

        Returns
        -------
        float
            The population growth factor.
        """
        return .5 * self.expected_offspring_per_pair * self.expected_mates_per_female

    def _dependency_graph(self):
        edge_list = []
        if self._haplotypes:
            edge_list.append(('proband\nhaplotypes', 'Mating\nregime'))
        if self._component_index is not None:
            for input_node in self._component_index._nodes:
                edge_list.append((input_node, 'Mating\nregime'))
        return edge_list

    @property
    def dependency_graph_edges(self):
        return self._dependency_graph()

    @property
    def dependency_graph(self):
        import networkx as nx
        G = nx.DiGraph()
        G.add_edges_from(self.dependency_graph_edges)
        pos = nx.nx_agraph.graphviz_layout(G, prog='dot')
        return (G,pos)

    def draw_dependency_graph(self, 
                              node_color='none', 
                              node_size = 1500, 
                              arrowsize = 7,
                              font_size=6, 
                              margins=.1, 
                              **kwargs):
        import networkx as nx
        G,pos = self.dependency_graph
        nx.draw_networkx(G,pos, 
                         node_color=node_color, 
                         node_size = node_size, 
                         font_size=font_size,
                         margins=margins,
                         arrowsize=arrowsize,
                         **kwargs)


class RandomMatingRegime(MatingRegime):
    """
    A mating regime that randomly pairs individuals and produces offspring with balanced numbers of males and females.

    Parameters
    ----------
    offspring_per_pair : xft.utils.VariableCount, optional
        Number of offspring produced per mating pair, by default xft.utils.ConstantCount(1)
    mates_per_female : xft.utils.VariableCount, optional
        Number of males that mate with each female, by default xft.utils.ConstantCount(1)
    female_offspring_per_pair : Union[str, xft.utils.VariableCount], optional
        The number of female offspring per mating pair. If "balanced", the number is balanced with
        the number of male offspring. By default, "balanced".
    sex_aware : bool, optional
        If True, randomly paired individuals are selected so that there is an equal number of males and females.
        Otherwise, random pairing is performed. By default, False.
    exhaustive : bool, optional
        If True, perform exhaustive enumeration of potential mates. If False, perform random sampling. By default, True.
    """
    def __init__(self, 
                 offspring_per_pair: xft.utils.VariableCount = xft.utils.ConstantCount(1),
                 mates_per_female: xft.utils.VariableCount =  xft.utils.ConstantCount(2),
                 female_offspring_per_pair: Union[str, xft.utils.VariableCount] = 'balanced', ## doesn't make total sense
                 sex_aware: bool = False,
                 exhaustive: bool = True,
                 ):
        super().__init__(self,
                         offspring_per_pair=offspring_per_pair,
                         mates_per_female=mates_per_female,
                         female_offspring_per_pair=female_offspring_per_pair,
                         sex_aware=sex_aware,
                         exhaustive=exhaustive,
                         component_index = None,
                         haplotypes = False,
                         )

    def mate(self, 
             haplotypes: xr.DataArray = None, 
             phenotypes: xr.DataArray = None,
             control: dict = None,
             ):
        """
        Mate individuals randomly with balanced numbers of males and females.

        Parameters
        ----------
        haplotypes : xr.DataArray, optional
            Array containing haplotypes, by default None
        phenotypes : xr.DataArray, optional
            Array containing phenotypes, by default None
        control : dict, optional
            Control dictionary, by default None

        Returns
        -------
        MateAssignment
            An object containing the maternal and paternal sample indices, the number of offspring per pair,
            and the number of female offspring per pair.
        """
        female_indices, male_indices = self.get_potential_mates(haplotypes, phenotypes)

        female_indices = np.random.permutation(female_indices)
        male_indices = np.random.permutation(male_indices)

        return self.enumerate_assignment(female_indices=female_indices,
                                         male_indices=male_indices,
                                         haplotypes=haplotypes,
                                         phenotypes=phenotypes,
                                         )

# mr =RandomMatingRegime()





class LinearAssortativeMatingRegime(MatingRegime):
    """
    A linear assortative mating regime that performs mate selection based on a specified component index. 
    Speifically, individuals are mated such that the cross-mate correlations across all specified components
    are equal to `r`. This reflects mating on a linear combination of phenotypes and does not generalize to
    many cross-mate correlation stuctures observed in practice, but is more efficient. 

    Parameters
    ----------
    component_index : xft.index.ComponentIndex
        The component index used to select mating pairs based on the correlation between the phenotype values.
    r : float, optional
        The correlation coefficient, a value between -1 and 1. Defaults to 0.
    offspring_per_pair : Union[int, xft.utils.VariableCount], optional
        The number of offspring per pair. If int, it will be converted to a ConstantCount object. Defaults to 1.
    mates_per_female : Union[int, xft.utils.VariableCount], optional
        The number of mates per female. If int, it will be converted to a ConstantCount object. Defaults to 1.
    female_offspring_per_pair : Union[str, int, xft.utils.VariableCount], optional
        The number of female offspring per mating pair. If 'balanced', the number of females is randomly selected 
        for each pair to balance the sex ratio. If int, it will be converted to a ConstantCount object. Defaults 
        to 'balanced'.
    sex_aware : bool, optional
        If True, only mating pairs with different sex are allowed. Defaults to False.
    exhaustive : bool, optional
        If True, all possible mating pairs will be enumerated. If False, pairs will be randomly selected.
        Defaults to True.
    
    Raises
    ------
    AssertionError
        If r is not between -1 and 1.
        If the correlation r is not feasible for the number of phenotypes in the component index.


    TODO: see also
    
    """
    def __init__(self, 
                 component_index: xft.index.ComponentIndex,
                 r: float = 0,
                 offspring_per_pair: Union[int, xft.utils.VariableCount] = xft.utils.ConstantCount(1),
                 mates_per_female: Union[int, xft.utils.VariableCount] =  xft.utils.ConstantCount(2),
                 female_offspring_per_pair: Union[str, int, xft.utils.VariableCount] = 'balanced', ## doesn't make total sense
                 sex_aware: bool = False,
                 exhaustive: bool = True,
                 ):
        super().__init__(self,
                         offspring_per_pair=offspring_per_pair,
                         mates_per_female=mates_per_female,
                         female_offspring_per_pair=female_offspring_per_pair,
                         sex_aware=sex_aware,
                         exhaustive=exhaustive,
                         component_index=component_index,
                         haplotypes=False,
                         )
        ## Linear regime attributes
        self.r = r
        self.component_index = component_index
        self.K = component_index.k_total
        assert (-1. <= r <= 1.), "Invalid correlation"
        if r > 1/self.K or r < -1/self.K:
            warnings.warn(f"Linear regime with {self.K} phenotypes only feasible for -1/{self.K} <= r <= 1/{self.K}")

    def mate(self, 
             haplotypes: xr.DataArray = None, 
             phenotypes: xr.DataArray = None,
             control: dict = None,
             ):
        female_indices, male_indices = self.get_potential_mates(haplotypes, phenotypes)

        sample_indexer = phenotypes.xft.get_sample_indexer()
        female_indexer = sample_indexer[female_indices] 
        male_indexer = sample_indexer[male_indices] 

        ## downsample for balance
        n_new = np.min([female_indexer.n, male_indexer.n])
        female_indexer = female_indexer.at_most(n_new)
        male_indexer = male_indexer.at_most(n_new)

        
        female_components = xft.utils.standardize_array(phenotypes.xft[female_indexer,
                                                        self.component_index].data)
        male_components = xft.utils.standardize_array(phenotypes.xft[male_indexer,
                                                      self.component_index].data)

        n=male_components.shape[0]
        sum_scaled_mate1 = np.sum(female_components, axis=1)
        sum_scaled_mate2 = np.sum(male_components, axis=1)
        within_cov1 = np.cov(female_components, rowvar=False)
        within_cov2 = np.cov(male_components, rowvar=False)
        cross_cov = self.K**2 * self.r
        R = np.sum(cross_cov)/ np.sqrt(np.sum(within_cov1)*(np.sum(within_cov2)))
        mating_score1 = sum_scaled_mate1 * np.sqrt(R) + np.sqrt(np.abs(1-R))*np.random.normal(0,np.std(sum_scaled_mate1),n)#sum_scaled_mate1.shape[0])
        mating_score2 = sum_scaled_mate2 * np.sqrt(R) + np.sqrt(np.abs(1-R))*np.random.normal(0,np.std(sum_scaled_mate2),n)#sum_scaled_mate2.shape[0])

        return self.enumerate_assignment(female_indices=female_indices[np.argsort(mating_score1)[:n_new]],
                                         male_indices=male_indices[np.argsort(mating_score2)[:n_new]],
                                         haplotypes=haplotypes,
                                         phenotypes=phenotypes,
                                         )



class BatchedMatingRegime(MatingRegime):
    """
    BatchedMatingRegime class that batches mating assignments, either for the sake of efficiency or to
    simulate stratification.

    Parameters
    ----------
    regime : MatingRegime
        The mating regime object.
    max_batch_size : int
        Maximum size of each batch.

    Attributes
    ----------
    regime : MatingRegime
        The mating regime object.
    max_batch_size : int
        Maximum size of each batch.

    Methods
    -------
    batch(haplotypes, phenotypes, control)
        Split samples into batches.
    mate(haplotypes, phenotypes, control)
        Generate mating assignments in batches.

    """
    def __init__(self, 
                 regime: MatingRegime, 
                 max_batch_size: int,
                 ):
        self.regime= regime
        self.max_batch_size = max_batch_size

    def batch(self, 
              haplotypes: xr.DataArray = None, 
              phenotypes: xr.DataArray = None,
              control: dict = None,
              ):
        """
        Split samples into batches.

        Parameters
        ----------
        haplotypes : xarray.DataArray, optional
            Haplotypes array.
        phenotypes : xarray.DataArray, optional
            Phenotypes array.
        control : dict, optional
            Control parameters.

        Returns
        -------
        batches : list
            List of batches of samples.
        num_batches : int
            Number of batches.

        """
        N = haplotypes.xft.n
        num_batches = math.ceil(N / self.max_batch_size)
        batches = [np.sort(x) for x in np.array_split(np.random.permutation(N), num_batches)]
        return (batches, num_batches)

    def mate(self,
             haplotypes: xr.DataArray = None,
             phenotypes: xr.DataArray = None,
             control: dict = None,
             ):
        """
        Generate mating assignments in batches and merge into single assignment object.

        Parameters
        ----------
        haplotypes : xarray.DataArray, optional
            Haplotypes array.
        phenotypes : xarray.DataArray, optional
            Phenotypes array.
        control : dict, optional
            Control parameters.

        Returns
        -------
        mate_assignments : MateAssignment
            Mating assignments.

        """
        batches, num_batches = self.batch(haplotypes,
                                          phenotypes,
                                          control)
        assignments = [self.regime.mate(haplotypes[batch],
                                        phenotypes[batch],
                                        control) for batch in batches]
        return MateAssignment.reduce_merge(assignments)




def _solve_qap_ls(Y, Z, R, nb_threads=6, time_limit=30, tolerance=1e-5, verbosity=1, time_between_displays=1):
    """
    Solves the Quadratic Assignment Problem (QAP) using the LocalSolver optimization solver.
    
    Parameters
    ----------
    Y : numpy.ndarray
        The first set of mate phenotypes.
    Z : numpy.ndarray
        The second set of mate phenotypes.
    R : numpy.ndarray
        The target cross-correlation matrix.
    nb_threads : int, optional
        The number of threads to be used for the optimization.
    time_limit : int, optional
        The amount of time (in seconds) to run the optimization passed to `.param.set_time_limit()`
    tolerance : float, optional
        The minimum threshold for the objective function to terminate optimization.
    verbosity : int, optional
        The amount of time (in seconds) to run the optimization.

    Returns
    -------
    P : numpy.ndarray
        A permutation matrix.
    """
    import localsolver
    n = Y.shape[0]
    # for later use as initial value
    tmp = np.argsort(np.apply_along_axis(np.mean, 1, Y))[
        np.argsort(np.argsort(np.apply_along_axis(np.mean, 1, Z)))]
    with localsolver.LocalSolver() as ls:
        ls.param.set_time_limit(int(time_limit))
        ls.param.set_nb_threads(int(nb_threads))
        ls.param.set_verbosity(verbosity)
        ls.param.set_time_between_displays(time_between_displays)

        model = ls.model
        # flows
        YY = Y @ Y.T / (n - 1)
        array_YY = model.array(model.array(YY[i, :]) for i in range(n))
        # distance
        ZZ = Z @ Z.T / (n - 1)
        # cost
        W = Y @ R @ Z.T / (n - 1)
        array_W = model.array(model.array(W[i, :]) for i in range(n))
        # permutation
        p = model.list(n)
        model.constraint(model.eq(model.count(p), n))
        # objective
        const = np.trace(W @ W.T)
        qobj = model.sum(
            model.at(array_YY, p[i], p[j]) * ZZ[i, j] for j in range(n) for i in range(n))
        lobj = model.sum(model.at(array_W, p[i], i) for i in range(n))
        obj = qobj - 2 * lobj + const
        model.minimize(obj)
        model.close()
        # set initial value of permutation
        p.value.clear()
        for pp in tmp:
            p.value.add(pp)
        # solve
        # ls.param.set_objective_threshold(0,tolerance)
        ls.param.set_verbosity(1)
        ls.solve()

        # solution
        P = np.array([p.value.get(i) for i in range(n)])

        return P

##TODO add seed
class GeneralAssortativeMatingRegime(MatingRegime):
    """
    A class that implements the general assortative mating regimes. I.e., matches two sets of individuals with
    K phenotypes to achieve an arbitrary K x K cross-mate cross-correlation structure.

    Parameters
    ----------
    component_index : xft.index.ComponentIndex
        An object containing information about the components.
    cross_corr : ndarray
        The cross-correlation matrix of size K x K.
    offspring_per_pair : Union[int, xft.utils.VariableCount], optional
        The number of offspring per mating pair. Default is 1.
    mates_per_female : Union[int, xft.utils.VariableCount], optional
        The number of mates for each female. Default is 2.
    female_offspring_per_pair : Union[str, int, xft.utils.VariableCount], optional
        The number of offspring per mating pair for females. Default is 'balanced'.
    sex_aware : bool, optional
        Whether to consider sex in mating pairs. Default is False.
    exhaustive : bool, optional
        Whether to enumerate all possible pairs. Default is True.
    control : dict, optional
        A dictionary of control parameters passed to LocalSolver. Defaults are as follows: 
        nb_threads=4, time_limit=120, tolerance=1e-5, verbosity=1, time_between_displays=15

    Attributes
    ----------
    cross_corr : ndarray
        The cross-correlation matrix of size K x K.
    component_index : xft.index.ComponentIndex
        An object containing information about the components.
    K : int
        The total number of components.

    Methods
    -------
    mate(haplotypes: xr.DataArray = None, phenotypes: xr.DataArray = None, control: dict = None) -> xft.mate.MateAssignment:
        Mate haplotypes and phenotypes based on the K-assortative mating regime.
    """

    def __init__(self,
                 component_index: xft.index.ComponentIndex,
                 cross_corr: NDArray,
                 offspring_per_pair: Union[int, xft.utils.VariableCount] = xft.utils.ConstantCount(
                     1),
                 mates_per_female: Union[int, xft.utils.VariableCount] = xft.utils.ConstantCount(
                     2),
                 # doesn't make total sense
                 female_offspring_per_pair: Union[str, int,
                                                  xft.utils.VariableCount] = 'balanced',
                 sex_aware: bool = False,
                 exhaustive: bool = True,
                 control: dict = {},

                 ):
        super().__init__(self,
                         offspring_per_pair=offspring_per_pair,
                         mates_per_female=mates_per_female,
                         female_offspring_per_pair=female_offspring_per_pair,
                         sex_aware=sex_aware,
                         exhaustive=exhaustive,
                         )
        # Linear regime attributes
        self.cross_corr = cross_corr
        self.component_index = component_index
        self.K = component_index.k_total
        self.control = dict(nb_threads=4, time_limit=120, tolerance=1e-5, verbosity=1, time_between_displays=15)
        self.control.update(control)

    def mate(self,
             haplotypes: xr.DataArray = None,
             phenotypes: xr.DataArray = None,
             control: dict = {},
             ):
        """
        Mate haplotypes and phenotypes based on the K-assortative mating regime.

        Parameters
        ----------
        haplotypes : xr.DataArray, optional
            The haplotype data to be mated. Default is None.
        phenotypes : xr.DataArray, optional
            The phenotype data to be mated. Default is None.

        Returns
        -------
        assignment : xft.mate.MateAssignment
            The assignment of haplotypes to parents.
        """
        female_indices, male_indices = self.get_potential_mates(
            haplotypes, phenotypes)

        sample_indexer = phenotypes.xft.get_sample_indexer()
        female_indexer = sample_indexer[female_indices]
        male_indexer = sample_indexer[male_indices]

        # downsample for balance
        n_new = np.min([female_indexer.n, male_indexer.n])
        female_indexer = female_indexer.at_most(n_new)
        male_indexer = male_indexer.at_most(n_new)

        female_components = xft.utils.standardize_array(phenotypes.xft[female_indexer,
                                                                       self.component_index].data)
        male_components = xft.utils.standardize_array(phenotypes.xft[male_indexer,
                                                                     self.component_index].data)

        perm = _solve_qap_ls(
            female_components, male_components, self.cross_corr, 
                nb_threads=self.control['nb_threads'],
                time_limit=self.control['time_limit'],
                tolerance=self.control['tolerance'],
                verbosity=self.control['verbosity'],
                time_between_displays=self.control['time_between_displays'],
                )

        return self.enumerate_assignment(female_indices=female_indices[perm],
                                         male_indices=male_indices,
                                         haplotypes=haplotypes,
                                         phenotypes=phenotypes,
                                         )



#     couplings <- rbind(mate1_inds[order(mating_score1)],
#                        mate2_inds[order(mating_score2)])
#     couplings <- rbind(mate1_inds[order(mating_score1)],
#                        mate2_inds[order(mating_score2)])
#     return(couplings)







#     def _mateFunction(self, 
#                       haplotypes: xr.DataArray = None, 
#                       phenotypes: xr.DataArray = None):

#         sample_indexer = haplotypes.xft.get_sample_indexer()

#         if self.sex_aware:
#             female_indices = haplotypes.sample[haplotypes.sex==0]
#             male_indices = haplotypes.sample[haplotypes.sex==1]
#             assert len(female_indices) == len(male_indices), "Unbalanced population"
#         elif not self.sex_aware:
#             permuted_indices = np.random.permutation(haplotypes.sample)
#             female_indices = np.sort(permuted_indices[0::2])
#             male_indices = np.sort(permuted_indices[1::2])

#         matings_per_female = self.mates_per_female.draw(len(female_indices))
#         n_matings = np.sum(matings_per_female)

#         offspring_per_mate_pair = self.offspring_per_pair.draw(n_matings)

#         female_mate_indices = xft.utils.exhaustive_enumerate(female_indices, matings_per_female)

#         if not isinstance(self.female_offspring_per_pair, str):
#             raise NotImplementedError
#         elif self.female_offspring_per_pair == 'balanced':
#                 n_females_per_pair = np.apply_along_axis(lambda x: np.random.randint(x+1), 0, offspring_per_mate_pair)
#         else:            
#             raise NotImplementedError
#         if self.exhaustive:
#             male_mate_indices = xft.utils.exhaustive_enumerate(male_indices, matings_per_female)
#         else:
#             male_mate_indices = np.random.choice(male_indices, n_matings)
#         # print(f"male_indices: {male_indices}, n_matings: {n_matings}, male_mate_indices: {male_mate_indices}")

#         return MateAssignment(generation = haplotypes.generation,
#                               maternal_sample_index = sample_indexer[female_mate_indices],
#                               paternal_sample_index = sample_indexer[male_mate_indices],
#                               previous_generation_sample_index = sample_indexer,
#                               n_offspring_per_pair = offspring_per_mate_pair,
#                               n_females_per_pair = n_females_per_pair,
#                               sex_aware = self.sex_aware,
#                               )




#     def _mateFunction(self, 
#                       haplotypes: xr.DataArray = None, 
#                       phenotypes: xr.DataArray = None):

#         sample_indexer = haplotypes.xft.get_sample_indexer()

#         if self.sex_aware:
#             female_indices = haplotypes.sample[haplotypes.sex==0]
#             male_indices = haplotypes.sample[haplotypes.sex==1]
#             assert len(female_indices) == len(male_indices), "Unbalanced population"
#         elif not self.sex_aware:
#             permuted_indices = np.random.permutation(haplotypes.sample)
#             female_indices = permuted_indices[0::2]
#             male_indices = permuted_indices[1::2]

#         matings_per_female = self.mates_per_female.draw(len(female_indices))
#         n_matings = np.sum(matings_per_female)

#         offspring_per_mate_pair = self.offspring_per_pair.draw(n_matings)


#         female_mate_indices = xft.utils.exhaustive_enumerate(female_indices, matings_per_female)

#         if not isinstance(self.female_offspring_per_pair, str):
#             raise NotImplementedError
#         elif self.female_offspring_per_pair == 'balanced':
#                 n_females_per_pair = np.apply_along_axis(lambda x: np.random.randint(x+1), 0, offspring_per_mate_pair)
#         else:            
#             raise NotImplementedError
#         if self.exhaustive:
#             male_mate_indices = xft.utils.exhaustive_enumerate(male_indices, matings_per_female)
#         else:
#             male_mate_indices = np.random.choice(male_indices, n_matings)
#         # print(f"male_indices: {male_indices}, n_matings: {n_matings}, male_mate_indices: {male_mate_indices}")
#     pass

# class KtraitAssortativeMatingRegime(MatingRegime):
#     ## TODO
#     pass

# class ChunkedMatingRegime(MatingRegime):
#     ## TODO
#     pass


#     #     if (not cp) and (not sa) and (opp==1) 

#     # def mate(self, 
#     #          phenotypes: Phenotypes = None,
#     #          pedigree: Pedigree = None,
#     #          ) -> MateAssignment:
#     #     raise NotImplementedError
# # class SuperClass:
# #     def __init__(self):
# #         pass
# #     def mate(self, *args):
# #         raise NotImplementedError

# # class SubClass(SuperClass):
# #     def __init__(self, params):
# #         self.params = params

# #     def mate(self):
# #         return self.params

# mate_linear_order <- function(yz, r) {
#     ## shuffle mates to avoid sex differences
#     n <- 2*(nrow(yz)%/%2)
#     K = ncol(yz)
#     permuted_inds <- sample.int(nrow(yz),n)
#     mate1_inds <- sort(permuted_inds[1:(n/2)])
#     mate2_inds <- sort(permuted_inds[(n/2+1):n])
#     scaled_mate1 <- (apply(yz[mate1_inds,,drop=FALSE ],2,scale))
#     scaled_mate2 <- (apply(yz[mate2_inds,,drop=FALSE ],2,scale))
#     sum_scaled_mate1 <- rowSums(scaled_mate1)
#     sum_scaled_mate2 <- rowSums(scaled_mate2)
#     within_cov1 <- cov(scaled_mate1)
#     within_cov2 <- cov(scaled_mate2)
#     cross_cov <- K^2 * r
#     R = sum(cross_cov)/ sqrt(sum(within_cov1)*(sum(within_cov2)))
#                                         #     print(R)
#     mating_score1 <- sum_scaled_mate1 * sqrt(R) + sqrt(abs(1-R))*rnorm(n/2,0,sd(sum_scaled_mate1))
#     mating_score2 <- sum_scaled_mate2 * sqrt(R) + sqrt(abs(1-R))*rnorm(n/2,0,sd(sum_scaled_mate2))
#     couplings <- rbind(mate1_inds[order(mating_score1)],
#                        mate2_inds[order(mating_score2)])
#     couplings <- rbind(mate1_inds[order(mating_score1)],
#                        mate2_inds[order(mating_score2)])
#     return(couplings)
# }